from pathlib import Path
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from collections import defaultdict
import lazy_dataset
from urllib.request import urlretrieve
from sed_scores_eval.base_modules.scores import (
    create_score_dataframe, validate_score_dataframe,
)
from sed_scores_eval.base_modules.detection import scores_to_event_list
from sed_scores_eval.base_modules.ground_truth import onset_offset_times_to_indices


def parse_inputs(scores, ground_truth, *, tagging=False):
    """read scores and ground_truth from files if string or path provided and
    validate audio ids

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.

    Returns:
        scores:
        ground_truth:
        audio_ids:

    """
    scores, audio_ids = parse_scores(scores)
    ground_truth = parse_ground_truth(ground_truth, tagging=tagging, audio_ids=audio_ids)
    return scores, ground_truth, audio_ids


def parse_scores(scores):
    if not isinstance(scores, (dict, str, Path, lazy_dataset.Dataset)):
        raise ValueError(
            f'scores must be dict, str, pathlib.Path or lazy_dataset.Dataset '
            f'but {type(scores)} was given.'
        )
    if isinstance(scores, (str, Path)):
        scores = Path(scores)
        scores = lazy_sed_scores_loader(scores)
    audio_ids = sorted(scores.keys())
    return scores, audio_ids


def parse_ground_truth(
        ground_truth, *,
        tagging=False, audio_ids=None, additional_ids_ok=False
):
    if not isinstance(ground_truth, (dict, str, Path)):
        raise ValueError(
            f'ground_truth must be dict, str or Path but {type(ground_truth)} '
            f'was given.'
        )
    if isinstance(ground_truth, (str, Path)):
        ground_truth = Path(ground_truth)
        if tagging:
            ground_truth, _ = read_ground_truth_tags(ground_truth)
        else:
            ground_truth = read_ground_truth_events(ground_truth)
    if not tagging:
        assert_non_connected_events(ground_truth)
    if audio_ids is not None:
        if additional_ids_ok:
            ground_truth = {key: ground_truth[key] for key in audio_ids}
        elif not (ground_truth.keys() == set(audio_ids)):
            raise ValueError(
                f'ground_truth audio ids do not match audio_ids. '
                f'Missing ids: {set(audio_ids) - ground_truth.keys()}. '
                f'Additional ids: {ground_truth.keys() - set(audio_ids)}.'
            )
    return ground_truth


def assert_non_connected_events(ground_truth):
    for clip_id, events in ground_truth.items():
        per_class_events = defaultdict(list)
        for event in events:
            per_class_events[event[2]].append(event)
        for event_class, class_events in per_class_events.items():
            class_events = sorted(class_events)
            current_offset = -1e6
            for event in class_events:
                assert event[0] > current_offset, f'Connected/overlapping {event[2]} events found for clip {clip_id}: {class_events}'
                current_offset = event[1]


def parse_audio_durations(audio_durations, *, audio_ids=None, additional_ids_ok=False):
    if not isinstance(audio_durations, (dict, str, Path)):
        raise ValueError(
            f'audio_durations must be dict, str or Path but '
            f'{type(audio_durations)} was given.'
        )
    if isinstance(audio_durations, (str, Path)):
        audio_durations = Path(audio_durations)
        assert audio_durations.is_file(), audio_durations
        audio_durations = read_audio_durations(audio_durations)
    if audio_ids is not None:
        if additional_ids_ok:
            audio_durations = {key: audio_durations[key] for key in audio_ids}
        elif not (audio_durations.keys() == set(audio_ids)):
            raise ValueError(
                f'audio_durations audio ids do not match audio_ids. '
                f'Missing ids: {set(audio_ids) - audio_durations.keys()}. '
                f'Additional ids: {audio_durations.keys() - set(audio_ids)}.'
            )
    return audio_durations


def write_sed_scores(scores, storage_path, *, timestamps=None, event_classes=None):
    """write sound event detection scores to tsv file

    Args:
        scores ((dict of) pandas.DataFrame): containing onset and offset times
            of a score window in first two columns followed by sed score
            columns for each event class. If dict keys are expected to be
            audio ids with corresponding data frames as values.
        storage_path (str or pathlib.Path): path to directory/file that is to be written
        timestamps (np.ndarray or list of float): optional list of timestamps
            to be compared with timestamps in scores DataFrame
        event_classes (list of str): optional list of event classes used to
            assert correct event labels in scores DataFrame

    """
    if isinstance(scores, dict):
        storage_path = Path(storage_path)
        storage_path.mkdir(exist_ok=True, parents=True)
        for audio_id, c_scores in scores.items():
            write_sed_scores(
                c_scores, storage_path / (audio_id + '.tsv'),
                timestamps=timestamps, event_classes=event_classes
            )
        return
    if not isinstance(scores, (np.ndarray, pd.DataFrame)):
        raise ValueError(
            f'scores must be np.ndarray or pd.DataFrame but {type(scores)}'
            f'was given.'
        )
    if isinstance(scores, np.ndarray):
        if timestamps is None:
            raise ValueError(
                f'timestamps must not be None if scores is np.ndarray'
            )
        if event_classes is None:
            raise ValueError(
                f'event_classes must not be None if scores is np.ndarray'
            )
        scores = create_score_dataframe(scores, timestamps, event_classes)
    validate_score_dataframe(scores, timestamps=timestamps, event_classes=event_classes)
    scores.to_csv(storage_path, sep='\t', index=False)


def read_sed_scores(filepath):
    """read scores from file(s)

    Args:
        filepath: path to tsv-file of a single utterance or to dir with tsv-files for the whole eval set

    Returns:

    """
    filepath = Path(filepath)
    if filepath.is_dir():
        scores = {}
        for file in sorted(filepath.iterdir()):
            if not file.is_file() or not file.name.endswith('.tsv'):
                raise ValueError(f'dir_path must only contain tsv files but contains {file}')
            scores[file.name[:-len('.tsv')]] = read_sed_scores(file)
        return scores
    scores = pd.read_csv(filepath, sep='\t')
    validate_score_dataframe(scores)
    return scores


def lazy_sed_scores_loader(dir_path):
    """lazy loader for sound event detection files in a directory. This is
    particularly useful if scores do not fit in memory for all audio files
    simultaneously.

    Args:
        dir_path (str or pathlib.Path): path to directory with sound event
            detection files
    """
    dir_path = Path(dir_path)
    if not dir_path.is_dir():
        raise NotADirectoryError(str(dir_path))
    score_files = {}
    for file in sorted(dir_path.iterdir()):
        if not file.is_file() or not file.name.endswith('.tsv'):
            raise ValueError(f'dir_path must only contain tsv files but contains {file}')
        score_files[file.name[:-len('.tsv')]] = str(file)
    scores = lazy_dataset.new(score_files)
    return scores.map(read_sed_scores)


def read_ground_truth_events(filepath):
    """read ground truth events from tsv file

    Args:
        filepath (str or pathlib.Path): path to file that is to be read.

    Returns:
        ground_truth (dict of lists of tuples): list of ground truth event
            tuples (onset, offset, event class) for each audio clip.

    """
    ground_truth = {}
    if isinstance(filepath, pd.DataFrame):
        file = filepath
    else:
        file = pd.read_csv(filepath, sep='\t')
    if not all([
        name in list(file.columns)
        for name in ['filename', 'onset', 'offset', 'event_label']
    ]):
        raise ValueError(
            f'ground_truth events file must contain columns "filename", '
            f'"onset", "offset" and "event_label" but only columns '
            f'{list(file.columns)} were found.'
        )
    for filename, onset, offset, event_label in zip(
        file['filename'], file['onset'], file['offset'], file['event_label']
    ):
        example_id = filename.rsplit('.', maxsplit=1)[0]
        if example_id not in ground_truth:
            ground_truth[example_id] = []
        if isinstance(event_label, str):
            assert len(event_label) > 0
            ground_truth[example_id].append([
                float(onset), float(offset), event_label
            ])
        else:
            # file without active events
            assert np.isnan(event_label), event_label
    return ground_truth


def read_ground_truth_tags(filepath):
    """read ground truth tags from tsv file

    Args:
        filepath (str or pathlib.Path): path to file that is to be read.

    Returns:
        tags (dict of lists): list of active events for each audio file.
        class_counts (dict of ints): number of files in which event_class is
            active for each event_class

    """
    tags = {}
    file = pd.read_csv(filepath, sep='\t')
    if 'filename' not in file.columns or (
            'event_label' not in file.columns
            and 'event_labels' not in file.columns
    ):
        raise ValueError(
            f'ground_truth tags file must contain columns "filename", '
            f'and "event_label" or "event_labels" but only columns '
            f'{list(file.columns)} were found.'
        )
    event_labels_key = "event_labels" if "event_labels" in file.columns else "event_label"
    class_counts = {}
    for filename, event_labels in zip(file['filename'], file[event_labels_key]):
        example_id = filename.rsplit('.', maxsplit=1)[0]
        if example_id not in tags:
            tags[example_id] = []
        if isinstance(event_labels, str):
            if event_labels_key == "event_labels":
                event_labels = event_labels.split(',')
            else:
                event_labels = [event_labels]
            for label in event_labels:
                tags[example_id].append(label)
                if label not in class_counts:
                    class_counts[label] = 0
                class_counts[label] += 1
        else:
            # file without active events
            assert np.isnan(event_labels), event_labels
    return tags, class_counts


def read_audio_durations(filepath):
    """read audio clip durations from tsv file

    Args:
        filepath (str or pathlib.Path): path to file that is to be read.

    Returns:
        audio_duration (dict of floats): audio duration in seconds for each
            audio file

    """
    audio_duration = {}
    file = pd.read_csv(filepath, sep='\t')
    assert [
        name in list(file.columns) for name in ['filename', 'duration']
    ], list(file.columns)
    for filename, duration in zip(file['filename'], file['duration']):
        example_id = filename.rsplit('.', maxsplit=1)[0]
        audio_duration[example_id] = float(duration)
    return audio_duration


def write_detection(
        scores, threshold, filepath, audio_format='wav',
):
    """perform thresholding of sound event detection scores and write detected
    events to tsv file

    Args:
        scores (dict of pandas.DataFrame): each DataFrame containing onset and
            offset times of a score window in first two columns followed by
            sed score columns for each event class. Dict keys have to be
            filenames without audio format ending.
        threshold ((dict of) float): threshold that is to be evaluated.
        filepath (str or pathlib.Path): path to file that is to be written/extended.
        audio_format: the audio format that is required to reconstruct the
            filename from audio ids/keys.

    """
    if not hasattr(scores, 'keys') or not callable(scores.keys):
        raise ValueError('scores must implement scores.keys()')
    keys = sorted(scores.keys())
    _, event_classes = validate_score_dataframe(scores[keys[0]])
    if isinstance(threshold, dict):
        threshold = [threshold[event_class] for event_class in event_classes]
        if not all([np.isscalar(thr) for thr in threshold]):
            raise ValueError('All values of thresholds dict must be scalars')
        threshold = np.asanyarray(threshold)
    elif not np.isscalar(threshold):
        raise ValueError(
            f'threshold must be (dict of) scalar(s) but {type(threshold)} '
            f'was given.'
        )
    filepath = Path(filepath)
    if not filepath.exists() or filepath.stat().st_size == 0:
        with Path(filepath).open('w') as fid:
            fid.write('filename\tonset\toffset\tevent_label\n')

    with filepath.open('a') as fid:
        event_lists = scores_to_event_list(scores, thresholds=threshold)
        for key, event_list in event_lists.items():
            for t_on, t_off, event_label in event_list:
                fid.write(
                    f'{key}.{audio_format}\t{t_on}\t{t_off}\t{event_label}\n')


def write_detections_for_multiple_thresholds(
        scores, thresholds, dir_path, audio_format='wav', score_transform=None,
        threshold_decimals=3,
):
    """writes a detection for multiple thresholds (operating points) as
    required by the psds_eval package (https://github.com/audioanalytic/psds_eval).
    This function is primarily used for testing purposes.

    Args:
        scores (dict of pandas.DataFrame): each DataFrame containing onset and
            offset times of a score window in first two columns followed by
            sed score columns for each event class. Dict keys have to be
            filenames without audio format ending.
        thresholds (np.array): an array of decision thresholds for each of
            which a detection file is written.
        dir_path (str or pathlib.Path): path to directory where to save
            detection files.
        audio_format: the audio format that is required to reconstruct the
            filename from audio ids/keys.
        score_transform:

    """
    if not hasattr(scores, 'keys') or not callable(scores.keys):
        raise ValueError('scores must implement scores.keys()')
    keys = sorted(scores.keys())
    thresholds = np.asanyarray(thresholds)
    if thresholds.ndim != 1:
        raise ValueError(
            f'thresholds must be a 1-dimensional array but has shape '
            f'{thresholds.shape}.'
        )
    assert np.all(np.abs(thresholds - np.round(thresholds, threshold_decimals)) < 1e-15), (threshold_decimals, thresholds)
    assert np.all(thresholds == np.unique(thresholds)), thresholds
    dir_path = Path(dir_path)
    dir_path.mkdir(parents=True, exist_ok=True)
    if score_transform is not None:
        if isinstance(score_transform, (str, Path)):
            score_transform = read_score_transform(score_transform)
        if not callable(score_transform):
            raise ValueError('score_transform must be callable.')
        if isinstance(scores, lazy_dataset.Dataset):
            scores = scores.map(score_transform)
        else:
            scores = {
                key: score_transform(scores_i)
                for key, scores_i in scores.items()
            }
    for key in keys:
        scores_i = scores[key]
        for threshold in thresholds:
            write_detection(
                {key: scores_i}, threshold,
                dir_path / '{:.Xf}.tsv'.replace('X', str(threshold_decimals)).format(threshold),
                audio_format=audio_format,
            )


def write_score_transform(
        scores, ground_truth, filepath,
        num_breakpoints=10, min_score=0., max_score=1.,
        classwise_transform=False,
):
    """compute and save a piecewise-linear score transform which is supposed
    to uniformly distribute scores from within ground truth events between 0
    and 1. This allows to obtain smoother PSD-ROC curve approximations when
    using the psds_eval package (https://github.com/audioanalytic/psds_eval)
    with linearly spaced thresholds.
    This function is primarily used for testing purposes.

    Args:
        scores (dict of pandas.DataFrames): score DataFrames for each audio
            clip of a data set. Each DataFrame contains onset and offset times
            of a score window  in first two columns followed by sed score
            columns for each event class.
        ground_truth (dict of lists of tuples): list of ground truth event
            tuples (onset, offset, event class) for each audio clip.
        filepath (str or pathlib.Path): path to file that is to be written.
        num_breakpoints: the number of breakpoints in the piecewise-linear
            transformation function.
        min_score: the first value (where y=x) in the transformation.
        max_score: the last value (where y=x) in the transformation.
        classwise_transform: If True, use separate transformations for scores
            from different event classes

    """
    scores, ground_truth, keys = parse_inputs(scores, ground_truth)
    scores_dict = {}
    event_classes = None
    for key in keys:
        scores_for_key = scores[key]
        timestamps, event_classes = validate_score_dataframe(
            scores_for_key, event_classes=event_classes)
        for (t_on, t_off, event_label) in ground_truth[key]:
            if event_label not in scores_dict:
                scores_dict[event_label] = []
            idx_on, idx_off = onset_offset_times_to_indices(
                onset_time=t_on, offset_time=t_off, timestamps=timestamps
            )
            scores_dict[event_label].append(scores_for_key[event_label][idx_on:idx_off])
    scores_dict = {
        event_class: np.concatenate(scores_dict[event_class])
        for event_class in event_classes
    }
    step = (max_score-min_score)/num_breakpoints
    output_scores = np.concatenate((
        [min_score],
        np.linspace(min_score, max_score, num_breakpoints, endpoint=False) + step/2,
        [max_score],
    ))
    output_scores = np.round(output_scores, decimals=12)
    score_transform = [output_scores]

    def get_breakpoints(score_arr):
        score_arr = np.sort(score_arr[score_arr > .0])
        breakpoints = score_arr[np.linspace(0, len(score_arr), num_breakpoints+1, dtype=int)[:-1]]
        return np.concatenate((
            [min_score], breakpoints, [max_score]
        ))

    if classwise_transform:
        for event_class in event_classes:
            score_transform.append(get_breakpoints(
                scores_dict[event_class]
            ))
    else:
        score_arr = np.concatenate([
            scores_dict[event_class] for event_class in event_classes])
        score_transform.extend(
            len(event_classes) * [get_breakpoints(score_arr)]
        )
    score_transform = pd.DataFrame(
        np.array(score_transform).T, columns=['y', *event_classes])
    score_transform.to_csv(filepath, sep='\t', index=False)


def read_score_transform(filepath):
    """read a piecewise linear score transform from tsv file

    Args:
        filepath: path to tsv file as written by write_score_transform

    Returns:
        score_transform: function which takes scores as pd.DataFrame and
            returns the transformed scores as pd.DataFrame

    """
    transform = pd.read_csv(filepath, sep='\t')
    column_names = list(transform.columns)
    assert len(column_names) > 1, column_names
    assert column_names[0] == 'y', column_names
    event_classes = column_names[1:]
    y = transform['y'].to_numpy()

    def score_transform(scores):
        validate_score_dataframe(scores, event_classes=event_classes)
        transformed_scores = [
            scores['onset'].to_numpy(), scores['offset'].to_numpy()
        ]
        for event_class in event_classes:
            x = transform[event_class].to_numpy()
            transformed_scores.append(interp1d(
                x, y, kind='linear',
            )(scores[event_class]))
        transformed_scores = pd.DataFrame(
            np.array(transformed_scores).T,
            columns=['onset', 'offset', *event_classes],
        )
        return transformed_scores

    return score_transform


def download_test_data():
    from sed_scores_eval import package_dir
    import zipfile
    tests_dir_path = package_dir / 'tests'
    if (tests_dir_path / 'data').exists():
        print('Test data already exists.')
        return
    print('Download test data')
    zip_file_path = tests_dir_path / 'data.zip'
    urlretrieve(
        'https://go.upb.de/sed_scores_eval_test_data',
        filename=str(zip_file_path)
    )
    with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
        zip_ref.extractall(tests_dir_path)
    zip_file_path.unlink()
    print('Download successful')



from pathlib import Path

import lazy_dataset
from functools import partial
from sed_scores_eval.base_modules.scores import validate_score_dataframe, create_score_dataframe
from sed_scores_eval.base_modules.cy_medfilt import cy_medfilt
from sed_scores_eval.base_modules.io import lazy_sed_scores_loader, write_sed_scores


def medfilt(scores, filter_length_in_sec, time_decimals=6, lazy=False, storage_dir=None):
    return apply_postprocessing(
        scores, cy_medfilt,
        filter_length_in_sec=filter_length_in_sec, time_decimals=time_decimals,
        lazy=lazy, storage_dir=storage_dir
    )


def apply_postprocessing(scores, postprocessing_fn, *, lazy=False, storage_dir=None, **kwargs):
    postprocessing_fn = partial(_postprocessing_fn_wrapper, postprocessing_fn=postprocessing_fn, **kwargs)
    if isinstance(scores, (str, Path)):
        scores = Path(scores)
        scores = lazy_sed_scores_loader(scores)
    elif lazy:
        scores = lazy_dataset.new(scores)
    if storage_dir is not None:
        storage_dir = Path(storage_dir)
        storage_dir.mkdir(exist_ok=True, parents=True)
    if lazy:
        scores = scores.map(postprocessing_fn)
        if storage_dir is not None:
            scores = lazy_dataset.zip(scores.keys(), scores).map(
                partial(_write_scores, storage_dir=storage_dir))
        return scores
    scores_processed = {}
    for audio_id in scores.keys():
        try:
            scores_processed[audio_id] = postprocessing_fn(scores[audio_id])
        except Exception as e:
            print(audio_id)
            raise e
        if storage_dir is not None:
            _write_scores((audio_id, scores_processed[audio_id]), storage_dir)
    return scores_processed


def _postprocessing_fn_wrapper(score_df, postprocessing_fn, **kwargs):
    timestamps, event_classes = validate_score_dataframe(score_df)
    score_arr = score_df[event_classes].to_numpy()
    score_arr_processed, timestamps_processed = postprocessing_fn(score_arr, timestamps, **kwargs)
    return create_score_dataframe(score_arr_processed, timestamps_processed, event_classes)


def _write_scores(audio_id_scores_df, storage_dir):
    audio_id, scores_df = audio_id_scores_df
    storage_path = Path(storage_dir) / f"{audio_id}.tsv"
    write_sed_scores(scores_df, storage_path)
    return scores_df



import numpy as np


def cummin(array):
    """cummulative minimum operation on 1d arrays

    Args:
        array:

    Returns:
        cummin_values (1d np.ndarray): cummin values
        cummin_indices (1d np.ndarray): indices of cummin values in input array

    >>> cummin(np.array([1.,3.,2.,0.,-1.,3,-2]))
    (array([ 1.,  1.,  1.,  0., -1., -1., -2.]), array([0, 0, 0, 3, 4, 4, 6]))
    >>> cummin(np.array([0.,0.,0.]))
    (array([0., 0., 0.]), array([0, 0, 0]))
    """
    cummin_values = np.minimum.accumulate(array)
    cummin_indices = np.concatenate((
        [0],
        np.arange(1, len(array))[cummin_values[1:] < cummin_values[:-1]]
    ))
    # _, cummin_indices = np.unique(cummin_values, return_index=True)
    return cummin_values, cummin_indices


def cummax(array):
    """cummulative maximum operation on 1d arrays

    Args:
        array:

    Returns:
        cummax_values (1d np.ndarray): cummax values
        cummax_indices (1d np.ndarray): indices of cummax values in input array

    >>> cummax(np.array([1.,3.,2.,0.,-1.,4.]))
    (array([1., 3., 3., 3., 3., 4.]), array([0, 1, 1, 1, 1, 5]))
    >>> cummax(np.array([0.,0.,0.]))
    (array([0., 0., 0.]), array([0, 0, 0]))
    """
    cummax_values = np.maximum.accumulate(array)
    cummax_indices = np.concatenate((
        [0],
        np.arange(1, len(array))[cummax_values[1:] > cummax_values[:-1]]
    ))
    # _, cummax_indices = np.unique(cummax_values, return_index=True)
    return cummax_values, cummax_indices


def get_first_index_where(array, criterion, value, *, axis=0):
    """get the first index where array fulfills a criterion w.r.t. value,
    where criterion may be "geq" (greater equal), "gt" (greater than),
    "leq" (less equal) or "lt" (less than). If criterion is met nowhere,
    function returns len(array)

    Args:
        array (1d np.ndarray):
        criterion (str):
        value (number):

    Returns:
        idx: first index where criterion is met or len(array) if criterion is
            never met.

    >>> arr = np.array([1,2,3,4,5])
    >>> get_first_index_where(arr, "geq", 3)
    2
    >>> get_first_index_where(arr, "gt", 3)
    3
    >>> get_first_index_where(arr, "gt", 5)
    5
    """
    if criterion == "geq":
        bool_idx = array >= value
    elif criterion == "gt":
        bool_idx = array > value
    elif criterion == "leq":
        bool_idx = array <= value
    elif criterion == "lt":
        bool_idx = array < value
    else:
        raise ValueError(f'Invalid criterion {criterion}')
    concat_shape = list(bool_idx.shape)
    concat_shape[axis] = 1
    bool_idx = np.concatenate((bool_idx, np.ones(concat_shape, dtype=bool)), axis=axis)
    return np.argmax(bool_idx, axis=axis)



# distutils: language = c++
#cython: language_level=3

import numpy as np
cimport numpy as np
cimport cython
from libcpp.vector cimport vector

cdef int searchsorted(double [:] sorted_list, double value, int init_idx):
    """Search position in a sorted list where value should be inserted.

    Args:
        sorted_list:
        value:
        init_idx: an index within the list from where to start searching

    Returns:

    """
    cdef int l = len(sorted_list)
    cdef int current_idx = min(init_idx, l-1)
    if sorted_list[current_idx] >= value:
        while current_idx > 0 and sorted_list[current_idx-1] >= value:
            current_idx -= 1
    else:
        while current_idx < l and sorted_list[current_idx] < value:
            current_idx += 1
    return current_idx

#@cython.boundscheck(False)  # Deactivate bounds checking
#@cython.wraparound(False)
def cy_medfilt(scores_in, timestamps_in, filter_length_in_sec=None, time_decimals=6):
    """

    Args:
        scores_in (2d np.ndarray): scores array with shape TxK for a clip with T scores segments and K classes
        timestamps_in: onset timestamps for each frame plus one more
            timestamp which is the final offset time.
        filter_length_in_sec: filter length in seconds
        time_decimals: the decimal precision to use w.r.t. timestamps.

    Returns:

    """

    assert filter_length_in_sec is not None, "You must provide filter_length_in_sec"

    if np.isscalar(filter_length_in_sec):
        if filter_length_in_sec == 0.:
            return scores_in, timestamps_in
        filter_length_in_sec = np.array(scores_in.shape[1] * [filter_length_in_sec], dtype=np.float64)


    scores_in = np.asanyarray(scores_in, dtype=np.float64)
    timestamps_in = np.asanyarray(timestamps_in, dtype=np.float64)
    filter_length_in_sec = np.asanyarray(filter_length_in_sec, dtype=np.float64)
    cdef double [:,:] overlap_breakpoint_timestamps = np.array([
        np.sort(
            np.round(np.concatenate((
                timestamps_in - fil_len / 2,
                timestamps_in + fil_len / 2,
            )), decimals=time_decimals + 2)
        )
        for fil_len in filter_length_in_sec
    ])
    # get breakpoints where the filter moves into or out of a segment

    cdef double [:] filter_length = filter_length_in_sec

    # Idea of median filter computation:
    # For a given time step, we compute the median as the center of a stacked
    # bar, where each segment, that overlaps with the filter at that time step,
    # is represented by a bar with height equal to the overlap with the filter.
    # Those bars are than stacked ordered by there corresponding score values
    # and the median value within the filter is than given as the center of the
    # stacked bar. This is done for all breakpoints and when there is a change
    # of median value between two breakpoints, the exact time point of median
    # value change can be computed by exploiting that between two breakpoints
    # the overlap/height of bars changes linearly. (The overlap/height of bar of
    # the segment the filter is moving out of, decreases linearly, whereas the
    # overlap/height of bar of the segment the filter is moving into,
    # increases linearly.)

    cdef int num_segments = scores_in.shape[0] + 2
    cdef int num_classes = scores_in.shape[1]
    cdef int num_breakpoints = overlap_breakpoint_timestamps.shape[1]
    cdef double [:,:] scores = np.concatenate((
        np.full_like(scores_in[:1], -np.inf),
        scores_in,
        np.full_like(scores_in[:1], -np.inf),
    ))
    cdef double [:] timestamps = np.concatenate(([-np.inf], timestamps_in, [np.inf]))
    cdef double eps = 10**(-time_decimals-1)

    cdef:
        int i, k, j, ranking_idx
        double t, t_diff, t_cp, current_cumsum, prev_cumsum, filter_length_k

    # for each class we track the current median score, the median bar lower
    # edge, which is the lower edge of the current median bar within the stacked
    # bar, and the bar upper edge, which is the upper edge of the current median
    # bar within the stacked bar.

    # track change_points where any of the classes change their median values
    cdef vector[double] change_point_timestamps = []
    cdef vector[int] change_point_class_labels = []
    cdef vector[double] change_point_class_scores = []
    cdef int relevant_segments_onset = 0
    cdef int relevant_segments_offset = 1
    cdef int num_relevant_segments = 1
    cdef int [:] ranking_indices = np.zeros(num_segments, dtype=np.int32)
    cdef double current_median_score = -np.inf
    cdef double current_median_score_range_low = 0
    cdef double current_median_score_range_up
    cdef double [:] scores_k
    cdef double [:] overlap_breakpoint_timestamps_k
    cdef double [:] relevant_scores = np.zeros(num_segments)
    cdef double [:] relevant_overlaps = np.zeros(num_segments)

    for k in range(num_classes):
        filter_length_k = filter_length[k]
        scores_k = scores[:, k]
        if filter_length_k == 0.:
            for i in range(1, num_segments):
                change_point_timestamps.push_back(timestamps[i])
                change_point_class_labels.push_back(k)
                change_point_class_scores.push_back(scores_k[i])
            continue
        current_median_score = -np.inf
        current_median_score_range_low = 0
        current_median_score_range_up = filter_length_k
        overlap_breakpoint_timestamps_k = overlap_breakpoint_timestamps[k]
        relevant_segments_onset = 0
        relevant_segments_offset = 1
        num_relevant_segments = 1
        relevant_scores[0] = -np.inf
        relevant_overlaps[0] = filter_length_k
        for i in range(1, num_breakpoints):
            t = overlap_breakpoint_timestamps_k[i]  # current breakpoint
            # get time difference to previous breakpoint, i.e., the time the filter
            # moved since last breakpoint
            t_diff = t - overlap_breakpoint_timestamps_k[i-1]
            if t_diff < eps:
                continue
            # get relevant segments range, i.e., segments the filter overlapped with
            # while moving from previous to current breakpoint and compute overlap
            # at current breakpoint
            if timestamps[relevant_segments_onset+1] < (t-filter_length_k/2 - eps):
                ranking_idx = searchsorted(
                    relevant_scores[:num_relevant_segments],
                    scores_k[relevant_segments_onset],
                    ranking_indices[relevant_segments_onset],
                )
                assert relevant_scores[ranking_idx] == scores_k[relevant_segments_onset]
                if relevant_overlaps[ranking_idx] < eps:
                    num_relevant_segments -= 1
                    for j in range(ranking_idx, num_relevant_segments):
                        relevant_scores[j] = relevant_scores[j+1]
                        relevant_overlaps[j] = relevant_overlaps[j+1]
                relevant_segments_onset += 1
            ranking_idx = searchsorted(
                relevant_scores[:num_relevant_segments],
                scores_k[relevant_segments_onset],
                ranking_indices[relevant_segments_onset],
            )
            assert relevant_scores[ranking_idx] == scores_k[relevant_segments_onset]
            relevant_overlaps[ranking_idx] -= t_diff
            ranking_indices[relevant_segments_onset] = ranking_idx
            if timestamps[relevant_segments_offset] < (t + filter_length_k/2 - eps):
                relevant_segments_offset += 1
                ranking_idx = searchsorted(
                    relevant_scores[:num_relevant_segments],
                    scores_k[relevant_segments_offset-1],
                    0,
                )
                ranking_indices[relevant_segments_offset - 1] = ranking_idx
                if ranking_idx == num_relevant_segments or relevant_scores[ranking_idx] != scores_k[relevant_segments_offset-1]:
                    num_relevant_segments += 1
                    for j in range(num_relevant_segments-1, ranking_idx, -1):
                        relevant_scores[j] = relevant_scores[j-1]
                        relevant_overlaps[j] = relevant_overlaps[j-1]
                    relevant_scores[ranking_idx] = scores_k[relevant_segments_offset-1]
                    relevant_overlaps[ranking_idx] = 0.
            else:
                ranking_idx = searchsorted(
                    relevant_scores[:num_relevant_segments],
                    scores_k[relevant_segments_offset-1],
                    ranking_indices[relevant_segments_offset - 1],
                )
            relevant_overlaps[ranking_idx] += t_diff
            if relevant_segments_offset == (relevant_segments_onset + 1):
                # if filter window lies completely within one segment than nothing changes
                continue
            #  update range of current median bar
            if scores_k[relevant_segments_onset] < current_median_score:
                current_median_score_range_low -= t_diff
                current_median_score_range_up -= t_diff
            elif scores_k[relevant_segments_onset] == current_median_score:
                current_median_score_range_up -= t_diff
            if scores_k[relevant_segments_offset-1] < current_median_score:
                current_median_score_range_low += t_diff
                current_median_score_range_up += t_diff
            elif scores_k[relevant_segments_offset-1] == current_median_score:
                current_median_score_range_up += t_diff
            # if center of stacked bar (filter_length_k / 2) is not within current
            # median bar range anymore, the median has changed while filter was
            # moving to the current breakpoint and we need to compute the exact
            # time point where the median value changed.
            if current_median_score_range_up < (filter_length_k/2 - eps):
                # change point
                current_cumsum = 0
                j = 0
                while current_cumsum < (filter_length_k/2 - eps):
                    prev_cumsum = current_cumsum
                    current_cumsum = current_cumsum + relevant_overlaps[j]
                    if current_cumsum > (current_median_score_range_up + eps):
                        t_cp = t - (filter_length_k/2 - prev_cumsum)
                        assert (relevant_scores[j] >= current_median_score), (relevant_scores[j], current_median_score, i, k, t)
                        if relevant_scores[j] > current_median_score:
                            change_point_timestamps.push_back(t_cp)
                            change_point_class_labels.push_back(k)
                            change_point_class_scores.push_back(relevant_scores[j])
                        if current_cumsum >= (filter_length_k/2 - eps):
                            current_median_score = relevant_scores[j]
                            current_median_score_range_up = current_cumsum
                            current_median_score_range_low = prev_cumsum
                    j += 1
            if current_median_score_range_low >= (filter_length_k/2 - eps):
                # change point
                current_cumsum = 0
                j = num_relevant_segments-1
                while current_cumsum <= (filter_length_k/2 - eps):
                    prev_cumsum = current_cumsum
                    current_cumsum = current_cumsum + relevant_overlaps[j]
                    if current_cumsum > (filter_length_k - current_median_score_range_up + eps):
                        t_cp = t - (filter_length_k/2 - prev_cumsum)
                        # assert (relevant_scores[j] <= current_median_scores), (relevant_scores[j], current_median_scores, i, k)
                        if relevant_scores[j] < current_median_score:
                            change_point_timestamps.push_back(t_cp)
                            change_point_class_labels.push_back(k)
                            change_point_class_scores.push_back(relevant_scores[j])
                        if current_cumsum >= (filter_length_k/2 - eps):
                            current_median_score = relevant_scores[j]
                            current_median_score_range_up = filter_length_k - prev_cumsum
                            current_median_score_range_low = filter_length_k - current_cumsum
                    j -= 1

    # wrap things up
    change_point_timestamps = np.round(np.array(change_point_timestamps), time_decimals)

    sort_idx = np.argsort(change_point_timestamps)
    change_point_timestamps_sorted = np.asarray(change_point_timestamps)[sort_idx]
    change_point_class_labels_sorted = np.asarray(change_point_class_labels)[sort_idx]
    change_point_class_scores_sorted = np.asarray(change_point_class_scores)[sort_idx]
    scores_filtered = []
    timestamps_filtered = []
    current_timestamp = -np.inf
    current_score_vec = scores.shape[1]*[-np.inf]
    for i, t in enumerate(change_point_timestamps_sorted):
        if t > (current_timestamp + eps):
            timestamps_filtered.append(t)
            scores_filtered.append(current_score_vec)
            current_score_vec = [*current_score_vec]
            current_timestamp = t
        current_score_vec[change_point_class_labels_sorted[i]] = change_point_class_scores_sorted[i]

    if len(scores_filtered) == 0:
        return np.full(scores_in.shape[1], -np.inf)[None], timestamps_in[[0,-1]]
    return np.array(scores_filtered)[1:], np.round(np.array(timestamps_filtered)+eps, time_decimals)



import matplotlib.pyplot as plt


def plot_psd_roc(
        psd_roc,
        *,
        dtc_threshold, gtc_threshold, cttc_threshold,
        alpha_ct, alpha_st, unit_of_time, max_efpr, psds,
        axes=None, filename=None,
        **kwargs
):
    """Shows (or saves) the PSD-ROC with optional standard deviation.

    This function is an adjustment from psds_eval.plot_psd_roc!

    When the plot is generated the area under PSD-ROC is highlighted.
    The plot is affected by the values used to compute the metric:
    max_efpr, alpha_ST and alpha_CT

    Args:
        psd_roc (tuple): The psd_roc that is to be plotted
        alpha_ct:
        alpha_st:
        unit_of_time:
        max_efpr:
        psds:
        axes (matplotlib.axes.Axes): matplotlib axes used for the plot
        filename (str): if provided a file will be saved with this name
        kwargs (dict): can set figsize
    """

    if not isinstance(psd_roc, tuple) or len(psd_roc) != 2:
        raise ValueError("The psd roc needs to be given as a tuple (etpr,efpr)")
    if axes is not None and not isinstance(axes, plt.Axes):
        raise ValueError("The give axes is not a matplotlib.axes.Axes")

    show = False
    if axes is None:
        fig = plt.figure(figsize=kwargs.get("figsize", (7, 7)))
        axes = fig.add_subplot()
        show = True

    axes.vlines(max_efpr, ymin=0, ymax=1.0, linestyles='dashed')
    etpr, efpr = psd_roc
    axes.step(efpr, etpr, 'b-', label='PSD-ROC', where="post")
    axes.fill_between(
        efpr, y1=etpr, y2=0, label="AUC", alpha=0.3, color="tab:blue",
        linewidth=3, step="post",
    )
    axes.set_xlim([0, max_efpr])
    axes.set_ylim([0, 1.0])
    axes.legend()
    axes.set_ylabel("eTPR")
    axes.set_xlabel(f"eFPR per {unit_of_time}")
    if cttc_threshold is None:
        cttc_threshold = "na"
    else:
        cttc_threshold = f"{cttc_threshold:.2f}"
    axes.set_title(
        f"PSDS: {psds:.5f}\n"
        f"dtc: {dtc_threshold:.2f}, "
        f"gtc: {gtc_threshold:.2f}, "
        f"cttc: {cttc_threshold}, "
        f"alpha_st: {alpha_st:.2f}, "
        f"alpha_ct: {alpha_ct:.2f}, "
        f"max_efpr: {max_efpr}"
    )
    axes.grid()
    if filename:
        plt.savefig(filename)
    if show:
        plt.show()
    else:
        plt.close()



import numpy as np
import pandas as pd
from sed_scores_eval.utils.array_ops import get_first_index_where


def create_score_dataframe(scores, timestamps, event_classes):
    """compose SED scores with event class labels into pandas.DataFrame with
    corresponding frame onset and offset times

    Args:
        scores (2d np.array): (T, K) SED scores for multiple event classes,
            with T being the number of frames and K being the number of event
            classes.
        timestamps (1d np.array): onset timestamps for each frame plus one more
            timestamp which is the final offset time.
        event_classes (list of str): list of event class names

    Returns: pandas.DataFrame with one row per frame where first and second
        column are the frame onset and offset time and the other columns the
        detection scores for the various event classes

    """
    if not isinstance(scores, (np.ndarray, list, tuple)):
        raise ValueError(
            f'scores must be np.ndarray, list or tuple but {type(scores)} '
            f'was given.'
        )
    scores = np.asanyarray(scores)
    if scores.ndim != 2:
        raise ValueError(
            f'scores must be two-dimensional but has shape {scores.shape}'
        )
    if not isinstance(timestamps, (np.ndarray, list, tuple)):
        raise ValueError(
            f'timestamps must be np.ndarray, list or tuple but '
            f'{type(timestamps)} was given.'
        )
    timestamps = np.asanyarray(timestamps)
    if timestamps.shape != (len(scores)+1,):
        raise ValueError(
            f'timestamps must be one-dimensional and have length '
            f'len(scores)+1 ({len(scores) + 1}) but timestamps with shape '
            f'{timestamps.shape} was given.'
        )

    if not isinstance(event_classes, (list, tuple)):
        raise ValueError(
            f'event_classes must be list or tuple but {type(event_classes)} '
            f'was given.'
        )
    if len(event_classes) != scores.shape[1]:
        raise ValueError(
            f'length of event_classes ({len(event_classes)}) does not match '
            f'scores.shape[1]. scores.shape: {scores.shape}.'
        )
    if not all([isinstance(c, (str, int)) for c in event_classes]):
        raise ValueError(
            f'All event classes must be either str or int but '
            f'event_classes={event_classes} was given.'
        )
    return pd.DataFrame(
        np.concatenate((
            timestamps[:-1, None], timestamps[1:, None], scores
        ), axis=1),
        columns=['onset', 'offset', *event_classes],
    )


def validate_score_dataframe(scores, timestamps=None, event_classes=None):
    """validate that scores is a pandas.DataFrame and has the correct format
    namely as provided by create_score_dataframe and return timestamps array
    and list of event class names.

    Args:
        scores: SED scores
        timestamps:
        event_classes (list of str): optional list of event classes used to
            assert correct event labels in scores DataFrame

    Returns:

    """
    if not isinstance(scores, pd.DataFrame):
        raise ValueError(
            f'scores must be pandas.DataFrame but {type(scores)} was given.')
    column_names = list(scores.columns)
    if (
        len(column_names) < 3
        or column_names[0] != 'onset'
        or column_names[1] != 'offset'
    ):
        raise ValueError(
            f'scores must contain at least 3 columns with first and second '
            f'column being frame onset and offset time, respectively, and '
            f'subsequent columns being score columns for various event '
            f'classes. However, provided columns are {column_names}.'
        )
    if event_classes is not None and column_names[2:] != event_classes:
        raise ValueError(
            f'column names {column_names[2:]} do not match the event class '
            f'names {event_classes}'
        )
    onset_times = scores['onset'].to_numpy()
    offset_times = scores['offset'].to_numpy()
    if (offset_times == onset_times).any():
        raise ValueError('Some frames have zero length.')
    if not (offset_times[:-1] == onset_times[1:]).all():
        raise ValueError(
            f'onset times must match offset times of the previous frame.'
        )
    timestamps_from_df = np.concatenate((onset_times, offset_times[-1:]))
    if timestamps is not None and not np.allclose(timestamps_from_df, timestamps):
        raise ValueError(
            f'timestamps from file {timestamps_from_df} do not match provided timestamps {timestamps}.'
        )
    event_classes = column_names[2:]
    for event_class in event_classes:
        assert "\\" not in event_class, ('class names must not contain "\\"', event_class)
    return timestamps_from_df, event_classes


def get_unique_thresholds(scores):
    """get thresholds lying between a unique score and next smaller unique score

    Args:
        scores (1d np.array): sed scores

    Returns:
        unique_thresholds (1d np.array): threshold values
        sort_indices (1d np.array): indices by which scores are sorted in
            ascending order
        unique_scores_indices (1d np.array): indices of the unique scores in
            the sorted scores array

    >>> score_arr = np.array([1,3,2,4,8,2])
    >>> get_unique_thresholds(score_arr)
    (array([-inf,  1.5,  2.5,  3.5,  6. ]), array([0, 2, 5, 1, 3, 4]), array([0, 1, 3, 4, 5]))
    """
    sort_indices = np.argsort(scores)
    sorted_scores = scores[sort_indices]
    unique_scores, unique_scores_indices = np.unique(
        sorted_scores, return_index=True
    )
    unique_thresholds = np.concatenate((
        [-np.inf], (unique_scores[1:] + unique_scores[:-1]) / 2
    ))
    return unique_thresholds, sort_indices, unique_scores_indices


def onset_offset_times_to_score_indices(onset_time, offset_time, timestamps):
    """

    Args:
        onset_time:
        offset_time:
        timestamps:

    Returns:
        onset_idx:
        offset_idx:

    """
    assert offset_time > onset_time, (onset_time, offset_time)
    # assert t_off <= timestamps[-1], (t_off, timestamps[-1])
    onset_idx = max(
        get_first_index_where(timestamps, 'gt', onset_time) - 1,
        0
    )
    assert timestamps[onset_idx] <= onset_time, (
        timestamps[onset_idx], onset_time)
    offset_idx = min(
        get_first_index_where(timestamps, 'geq', offset_time),
        len(timestamps) - 1,
    )
    assert offset_idx > onset_idx, (onset_idx, offset_idx)
    # assert timestamps[onset_idx] <= onset_time < timestamps[onset_idx+1]
    # assert timestamps[offset_idx-1] < offset_time <= timestamps[onset_idx+1]
    return onset_idx, offset_idx



# distutils: language = c++
#cython: language_level=3

import numpy as np
cimport numpy as np
cimport cython
from numpy.math cimport INFINITY


def onset_offset_curves(scores_in, timestamps_in, change_point_candidates_in=None):
    """get onset and offset times of event detections for various decision
    thresholds. Here, the number of event detections is given by the number of
    local maxima in the score signal with events being spawned when the
    decision threshold falls below the local maximum. However, usually only a
    subset of these events is active simultanously while others are inactive,
    e.g., because a certain threshold does not yet fall below all local
    maximums. For inactive events we return offset_time = onset_time. Further,
    when the decision threshold falls below a local minimum, two separate
    events merge into a single event. In this case, we keep the earlier event
    active with corresponding onset and offset times, while the later event is
    set inactive with offset_time = onset_time.

    Args:
        scores_in (1d np.ndarray): SED scores for a single event class
        timestamps_in (1d np.ndarray): onset timestamps for each score plus one more
            timestamp which is the final offset time.
        change_point_candidates_in (1d np.ndarray):

    Returns:
        scores_unique (1d np.ndarray): unique and sorted score array
        onset_times (2d np.ndarray): onset times for each possible event at
            each decsion threshold that falls below one of the unique scores.
            Shape is (number of unique scores, number of events/local maximums).
        offset_times (2d np.ndarray): offset times for each possible event at
            each decsion threshold that falls below one of the unique scores.
            Shape is (number of unique scores, number of events/local maximums).

    >>> y = np.array([.4,1.,.6,.6,1.,1.,.4])
    >>> ts = np.linspace(0., len(y)*.2, len(y) + 1)  # each score has width of 200ms
    >>> y, t_on, t_off = onset_offset_curves(y, ts)
    >>> y
    array([0.4, 0.6, 1. ])
    >>> np.stack((t_on, t_off), axis=-1)
    array([[[0. , 1.4],
            [0. , 0. ]],
    <BLANKLINE>
           [[0.2, 1.2],
            [0. , 0. ]],
    <BLANKLINE>
           [[0.2, 0.4],
            [0.8, 1.2]]])
    """
    scores_in = np.asanyarray(scores_in, dtype=np.float64)
    if not scores_in.ndim == 1:
        raise ValueError(
            f'scores must be 1-dimensional array of single class SED scores, '
            f'but array of shape {scores_in.shape} was given.'
        )
    timestamps_in = np.asanyarray(timestamps_in, dtype=np.float64)
    if not timestamps_in.ndim == 1 or len(timestamps_in) != (len(scores_in) + 1):
        raise ValueError(
            f'timestamps must be 1-dimensional array of length(len(scores) + 1), '
            f'but array of shape {timestamps_in.shape} was given.'
        )

    if change_point_candidates_in is None:
        change_point_candidates_in = np.unique(scores_in)
    change_point_candidates_in = np.asanyarray(change_point_candidates_in, dtype=np.float64)

    cdef double [:] scores = scores_in
    cdef double [:] timestamps = timestamps_in
    cdef double [:] change_point_candidates = change_point_candidates_in

    cdef int num_segments = len(scores)
    cdef int num_candidates = len(change_point_candidates)
    cdef int max_events = 0

    cdef double [:,:] onset_times = np.zeros((len(change_point_candidates), (num_segments+1)//2))
    cdef double [:,:] offset_times = np.zeros_like(onset_times)

    cdef:
        int i, t, event_idx
        double threshold

    for i in range(num_candidates):
        threshold = change_point_candidates[i]
        event_idx = 0
        is_on = False
        for t in range(num_segments):
            if is_on:
                if scores[t] < threshold or scores[t] == -INFINITY:
                    offset_times[i, event_idx] = timestamps[t]
                    event_idx += 1
                    is_on = False
            else:
                if scores[t] >= threshold and scores[t] > -INFINITY:
                    onset_times[i, event_idx] = timestamps[t]
                    is_on = True
        if is_on:
            offset_times[i, event_idx] = timestamps[num_segments]
            event_idx += 1
        max_events = max(max_events, event_idx)
    return np.array(change_point_candidates), np.array(onset_times[:, :max_events]), np.array(offset_times[:, :max_events])



import numpy as np
from scipy.interpolate import interp1d
from sed_scores_eval.utils.array_ops import cummax, get_first_index_where
from sed_scores_eval.utils import parallel
from sed_scores_eval.base_modules.curves import xsort, staircase_auc
from sed_scores_eval.base_modules.bootstrap import bootstrap_from_deltas
from sed_scores_eval.base_modules.io import parse_audio_durations
from sed_scores_eval.intersection_based.intermediate_statistics import intermediate_statistics_deltas, accumulated_intermediate_statistics

seconds_per_unit_of_time = {
    'second': 1.,
    'minute': 60.,
    'hour': 3600.,
}


def psds(
        scores, ground_truth, audio_durations, *, deltas=None,
        dtc_threshold, gtc_threshold, cttc_threshold=None,
        alpha_ct=.0, alpha_st=.0, unit_of_time='hour', max_efpr=100.,
        time_decimals=6, num_jobs=1,
):
    """Computes Polyphonic Sound Detection Score (PSDS) [1] using the exact
    and efficient computation approach proposed in [2].

    [1] C. Bilen, G. Ferroni, F. Tuveri, J. Azcarreta and S. Krstulovic,
    "A Framework for the Robust Evaluation of Sound Event Detection",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2020, pp. 61–65

    [2] J.Ebbers, R.Serizel, and R.Haeb-Umbach
    "Threshold-Independent Evaluation of Sound Event Detection Scores",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2022

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        audio_durations: The duration of each audio file in the evaluation set.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        dtc_threshold (float): detection tolerance criterion threshold
        gtc_threshold (float): ground truth intersection criterion threshold
        cttc_threshold (float): cross trigger tolerance criterion threshold
        alpha_ct (float): parameter for penalizing cross triggers.
            More specifically, it is the weight of the cross trigger rate
            (averaged over all other classes) that is added to the False
            Positive Rate (FPR) yielding the effective FPR (eFPR). Default is 0.
        alpha_st (float): parameter for penalizing instability across classes.
            More specifically, it is the weight of the standard deviation of
            the per-class ROCs, that is subtracted from the mean of the
            per-class ROCs. Default is 0.
        unit_of_time (str): the unit of time \in {second, minute, hour} to be
            used for computation of the eFPR (which is defined as rate per unit
            of time). Default is hour.
        max_efpr (float): the maximum eFPR for which the system is evaluated,
            i.e., until where the Area under PSD ROC Curve is computed.
            Default is 100.
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, e.g., a detection with an ground truth intersection
            exactly matching the DTC, may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        psds (float): Polyphonic Sound Detection Score (PSDS), i.e., the area
            under the PSD ROC Curve up to max_efpr normalized by max_efpr.
        single_class_psds (dict):
        psd_roc (tuple of 1d np.ndarrays): tuple of effective True Positive
            Rates and effective False Positive Rates.
        single_class_psd_rocs (dict of tuples of 1d np.ndarrays):
            tuple of True Positive Rates and effective False Positive Rates
            for each event class.

    """
    if isinstance(scores, (list, tuple)) or isinstance(deltas, (list, tuple)):
        # batch input
        batch_size = [len(v) for v in [scores, deltas] if v is not None][0]
        return list(zip(*parallel.map(
            (scores, deltas), arg_keys=('scores', 'deltas'),
            func=psds, max_jobs=num_jobs,
            ground_truth=ground_truth, audio_durations=audio_durations,
            dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold, alpha_ct=alpha_ct,
            alpha_st=alpha_st, unit_of_time=unit_of_time, max_efpr=max_efpr,
            time_decimals=time_decimals,
            num_jobs=num_jobs//batch_size,
        )))
    (effective_tp_rate, effective_fp_rate), single_class_psd_rocs = psd_roc(
        scores=scores, ground_truth=ground_truth,
        audio_durations=audio_durations, deltas=deltas,
        dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
        cttc_threshold=cttc_threshold, alpha_ct=alpha_ct, alpha_st=alpha_st,
        unit_of_time=unit_of_time, max_efpr=max_efpr,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    psds_value = psds_from_psd_roc(
        effective_tp_rate, effective_fp_rate, max_efpr)
    single_class_psds = {
        class_name: psds_from_psd_roc(tpr, efpr, max_efpr)
        for class_name, (tpr, efpr, _) in single_class_psd_rocs.items()
    }
    return (
        psds_value, single_class_psds,
        (effective_tp_rate, effective_fp_rate),
        single_class_psd_rocs
    )


def psd_roc(
        scores, ground_truth, audio_durations, *, deltas=None,
        dtc_threshold, gtc_threshold, cttc_threshold=None,
        alpha_ct=.0, alpha_st=.0, unit_of_time='hour', max_efpr=100.,
        time_decimals=6, num_jobs=1,
):
    """Computes Polyphonic Sound Detection ROC (PSD ROC) [1] using the exact
    and efficient computation approach proposed in [2].

    [1] C. Bilen, G. Ferroni, F. Tuveri, J. Azcarreta and S. Krstulovic,
    "A Framework for the Robust Evaluation of Sound Event Detection",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2020, pp. 61–65

    [2] J.Ebbers, R.Serizel, and R.Haeb-Umbach
    "Threshold-Independent Evaluation of Sound Event Detection Scores",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2022

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        audio_durations: The duration of each audio file in the evaluation set.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        dtc_threshold (float): detection tolerance criterion threshold
        gtc_threshold (float): ground truth intersection criterion threshold
        cttc_threshold (float): cross trigger tolerance criterion threshold
        alpha_ct (float): parameter for penalizing cross triggers.
            More specifically, it is the weight of the cross trigger rate
            (averaged over all other classes) that is added to the False
            Positive Rate (FPR) yielding the effective FPR (eFPR). Default is 0.
        alpha_st (float): parameter for penalizing instability across classes.
            More specifically, it is the weight of the standard deviation of
            the per-class ROCs, that is subtracted from the mean of the
            per-class ROCs. Default is 0.
        unit_of_time (str): the unit of time \in {second, minute, hour} to be
            used for computation of the eFPR (which is defined as rate per unit
            of time). Default is hour.
        max_efpr (float): the maximum eFPR for which the system is evaluated,
            i.e., until where the Area under PSD ROC Curve is computed.
            Default is 100.
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, e.g., a detection with an ground truth intersection
            exactly matching the DTC, may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        etpr (1d np.ndarray): effective True Positive Rates.
        efpr (1d np.ndarray): effective False Positive Rates.
        single_class_psd_rocs (dict of tuples of 1d np.ndarrays):
            tuple of True Positive Rates and effective False Positive Rates
            for each event class.

    """
    if alpha_ct == 0.:
        if cttc_threshold is not None:
            raise ValueError(
                'cttc_threshold has been provided but alpha_ct is 0.'
            )
    else:
        if cttc_threshold is None:
            raise ValueError(
                'alpha_ct is not 0 but no cttc_threshold has been provided.'
            )

    intermediate_statistics, audio_ids = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
        cttc_threshold=cttc_threshold,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )

    audio_durations = parse_audio_durations(
        audio_durations, audio_ids=audio_ids, additional_ids_ok=(deltas is not None))
    dataset_duration = sum(audio_durations.values())

    single_class_psd_rocs = _single_class_roc_from_intermediate_statistics(
        intermediate_statistics,
        dataset_duration=dataset_duration,
        alpha_ct=alpha_ct, unit_of_time=unit_of_time, max_efpr=max_efpr,
    )
    effective_tp_rate, overall_effective_fp_rates = multi_class_psd_roc_from_single_class_psd_rocs(single_class_psd_rocs, alpha_st=alpha_st, max_efpr=max_efpr)
    return (effective_tp_rate, overall_effective_fp_rates), single_class_psd_rocs


def psds_from_psd_roc(tpr, fpr, max_fpr):
    return staircase_auc(tpr, fpr, max_x=max_fpr) / max_fpr


def bootstrapped_psds(
        scores, ground_truth, audio_durations, *, deltas=None,
        dtc_threshold, gtc_threshold, cttc_threshold=None,
        alpha_ct=.0, alpha_st=.0, unit_of_time='hour', max_efpr=100.,
        time_decimals=6, n_folds=5, n_iterations=20, num_jobs=1,
):
    """

    Args:
        scores:
        ground_truth:
        audio_durations:
        deltas:
        dtc_threshold:
        gtc_threshold:
        cttc_threshold:
        alpha_ct:
        alpha_st:
        unit_of_time:
        max_efpr:
        time_decimals:
        n_folds:
        n_iterations:
        num_jobs:

    Returns:

    """
    if isinstance(scores, (list, tuple)) or isinstance(deltas, (list, tuple)):
        # batch input
        batch_size = [len(v) for v in [scores, deltas] if v is not None][0]
        return list(zip(*parallel.map(
            (scores, deltas), arg_keys=('scores', 'deltas'),
            func=bootstrapped_psds, max_jobs=num_jobs,
            ground_truth=ground_truth, audio_durations=audio_durations,
            dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold, alpha_ct=alpha_ct,
            alpha_st=alpha_st, unit_of_time=unit_of_time, max_efpr=max_efpr,
            time_decimals=time_decimals,
            n_folds=n_folds, n_iterations=n_iterations,
            num_jobs=num_jobs//batch_size,
        )))
    if deltas is None:
        deltas = intermediate_statistics_deltas(
            scores=scores, ground_truth=ground_truth,
            dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold,
            time_decimals=time_decimals, num_jobs=num_jobs,
        )
    return bootstrap_from_deltas(
        psds, deltas,
        n_folds=n_folds, n_iterations=n_iterations, num_jobs=num_jobs,
        scores=None, ground_truth=ground_truth, audio_durations=audio_durations,
        dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
        cttc_threshold=cttc_threshold, alpha_ct=alpha_ct, alpha_st=alpha_st,
        unit_of_time=unit_of_time, max_efpr=max_efpr
    )


def _single_class_roc_from_intermediate_statistics(
        scores_intermediate_statistics, dataset_duration,
        alpha_ct=.0, unit_of_time='hour', max_efpr=100.,
):
    """

    Args:
        scores_intermediate_statistics ((dict of) tuple): tuple of scores array and
            dict of intermediate_statistics with the following key value pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'cts' (dict of 1d np.ndarray): cross triggers with each of the
                other classes for each score
             'n_ref' (int): number of ground truth events
             't_ref_other' (dict of float): total ground truth event durations
                for each of the other classes.
            If dict input is provided, keys are expected to be class names with
            corresponding scores/intermediate_statistics as values.
        dataset_duration (float): total dataset duration, i.e., the sum of the
            individual file durations.
        alpha_ct (float): parameter for penalizing cross triggers.
            More specifically, it is the weight of the cross trigger rate
            (averaged over all other classes) that is added to the False
            Positive Rate (FPR) yielding the effective FPR (eFPR).
        unit_of_time (str): the unit of time \in {second, minute, hour} to be
            used for computation of the eFPR (which is defined as rate per unit
            of time). Default is hour.
        max_efpr (float): the maximum eFPR for which the system is evaluated,
            i.e., until where the Area under PSD ROC Curve is computed.

    Returns:
        tp_ratio (1d np.ndarray): True Positive Ratios
        effective_fp_rate (1d np.ndarray): effective False Positive Rates
        scores (1d np.ndarray): corresponding scores that the decision
            threshold has to fall below.

    """
    if unit_of_time not in seconds_per_unit_of_time:
        raise ValueError(
            f'Invalid unit_of_time {unit_of_time}. Valid units are '
            f'{", ".join(list(seconds_per_unit_of_time.keys()))}.'
        )

    if isinstance(scores_intermediate_statistics, dict):
        return {
            cls: _single_class_roc_from_intermediate_statistics(
                scores_stats,
                dataset_duration=dataset_duration,
                alpha_ct=alpha_ct,
                unit_of_time=unit_of_time,
                max_efpr=max_efpr,
            ) for cls, scores_stats in (
                scores_intermediate_statistics.items())
        }

    scores, stats = scores_intermediate_statistics

    tp_ratio = stats['tps'] / max(stats['n_ref'], 1)
    fp_rate = stats['fps'] / dataset_duration
    if alpha_ct == .0:
        effective_fp_rate = fp_rate
    else:
        assert stats['cts'].keys() == stats['t_ref_other'].keys(), (
            stats['cts'].keys(), stats['t_ref_other'].keys())
        assert all([value.shape == (len(scores),) for value in stats['cts'].values()]), (
            [value.shape for value in stats['cts'].values()], len(scores))
        ct_rates = [
            stats['cts'][class_name] / max(stats['t_ref_other'][class_name], 1e-12)
            for class_name in stats['cts']
        ]
        effective_fp_rate = fp_rate + alpha_ct * np.mean(ct_rates, axis=0)
    effective_fp_rate = effective_fp_rate * seconds_per_unit_of_time[unit_of_time]
    return _unique_cummax_sort(
        tp_ratio, effective_fp_rate, scores, max_efpr=max_efpr
    )


def multi_class_psd_roc_from_single_class_psd_rocs(single_class_psd_rocs, alpha_st, max_efpr):
    tp_ratios, efp_rates, *_ = list(zip(*single_class_psd_rocs.values()))
    overall_effective_fp_rates = np.unique(np.sort(np.concatenate(efp_rates)))
    interpolated_tp_ratios = []
    for tpr, efpr in zip(tp_ratios, efp_rates):
        if len(tpr) == 1:
            # interp1d expects at least length of 2, which, however, isn't
            # necessary with bounds_error=False and fill_values. Therefore,
            # simply repeat arrays if length == 1
            tpr = tpr.repeat(2)
            efpr = efpr.repeat(2)
        interpolated_tp_ratios.append(
            interp1d(
                efpr, tpr, kind='previous',
                bounds_error=False, fill_value=(0, tpr[-1])
            )(overall_effective_fp_rates)
        )
    mu_tp = np.mean(interpolated_tp_ratios, axis=0)
    sigma_tp = np.std(interpolated_tp_ratios, axis=0)
    effective_tp_rate = mu_tp - alpha_st * sigma_tp
    effective_tp_rate = np.maximum(effective_tp_rate, 0.)
    if max_efpr is not None:
        effective_tp_rate = np.concatenate((
            effective_tp_rate, [effective_tp_rate[-1]]
        ))
        overall_effective_fp_rates = np.concatenate((
            overall_effective_fp_rates, [max_efpr]
        ))
    return effective_tp_rate, overall_effective_fp_rates


def _unique_cummax_sort(tp_ratio, effective_fp_rate, *other, max_efpr=None):
    tp_ratio, effective_fp_rate, *other = xsort(tp_ratio, effective_fp_rate, *other)
    cummax_indices = cummax(tp_ratio)[1]
    tp_ratio, effective_fp_rate, *other = [values[cummax_indices] for values in [tp_ratio, effective_fp_rate, *other]]
    effective_fp_rate, unique_efpr_indices = np.unique(
        effective_fp_rate[::-1], return_index=True)
    unique_efpr_indices = - 1 - unique_efpr_indices
    tp_ratio, *other = [values[unique_efpr_indices] for values in [tp_ratio, *other]]
    if max_efpr is not None:
        cutoff_idx = get_first_index_where(effective_fp_rate, "gt", max_efpr)
        tp_ratio, effective_fp_rate, *other = [values[:cutoff_idx] for values in [tp_ratio, effective_fp_rate, *other]]
    return (tp_ratio, effective_fp_rate, *other)



from typing import Optional


def flatten(d, sep: Optional[str] = '.', *, flat_type=dict):
    """
    Flatten a nested `dict` using a specific separator.

    !!! Copied from https://github.com/fgnt/paderbox/blob/master/paderbox/utils/nested.py !!!

    Args:
        sep: When `None`, return `dict` with `tuple` keys (guarantees inversion
                of flatten) else join the keys with sep
        flat_type:  Allow other mappings instead of `flat_type` to be
                flattened, e.g. using an isinstance check.

    import collections
    flat_type=collections.abc.MutableMapping

    >>> d_in = {'a': 1, 'c': {'a': 2, 'b': {'x': 5, 'y' : 10}}, 'd': [1, 2, 3]}
    >>> d = flatten(d_in)
    >>> for k, v in d.items(): print(k, v)
    a 1
    c.a 2
    c.b.x 5
    c.b.y 10
    d [1, 2, 3]
    >>> d = flatten(d_in, sep='_')
    >>> for k, v in d.items(): print(k, v)
    a 1
    c_a 2
    c_b_x 5
    c_b_y 10
    d [1, 2, 3]
    """

    # https://stackoverflow.com/a/6027615/5766934

    # {k: v for k, v in d.items()}

    def inner(d, parent_key):
        items = {}
        for k, v in d.items():
            new_key = parent_key + (k,)
            if isinstance(v, flat_type) and v:
                items.update(inner(v, new_key))
            else:
                items[new_key] = v
        return items

    items = inner(d, ())
    if sep is None:
        return items
    else:
        return {
            sep.join(k): v for k, v in items.items()
        }


def deflatten(d: dict, sep: Optional[str] = '.', maxdepth: int = -1):
    """
    Build a nested `dict` from a flat dict respecting a separator.

    !!! Copied from https://github.com/fgnt/paderbox/blob/master/paderbox/utils/nested.py !!!

    Args:
        d: Flattened `dict` to reconstruct a `nested` dict from
        sep: The separator used in the keys of `d`. If `None`, `d.keys()` should
            only contain `tuple`s.
        maxdepth: Maximum depth until wich nested conversion is performed

    >>> d_in = {'a': 1, 'c': {'a': 2, 'b': {'x': 5, 'y' : 10}}, 'd': [1, 2, 3]}
    >>> d = flatten(d_in)
    >>> for k, v in d.items(): print(k, v)
    a 1
    c.a 2
    c.b.x 5
    c.b.y 10
    d [1, 2, 3]
    >>> deflatten(d)
    {'a': 1, 'c': {'a': 2, 'b': {'x': 5, 'y': 10}}, 'd': [1, 2, 3]}
    >>> deflatten(d, maxdepth=1)
    {'a': 1, 'c': {'a': 2, 'b.x': 5, 'b.y': 10}, 'd': [1, 2, 3]}
    >>> deflatten(d, maxdepth=0)
    {'a': 1, 'c.a': 2, 'c.b.x': 5, 'c.b.y': 10, 'd': [1, 2, 3]}
    >>> d = flatten(d_in, sep='_')
    >>> for k, v in d.items(): print(k, v)
    a 1
    c_a 2
    c_b_x 5
    c_b_y 10
    d [1, 2, 3]
    >>> deflatten(d, sep='_')
    {'a': 1, 'c': {'a': 2, 'b': {'x': 5, 'y': 10}}, 'd': [1, 2, 3]}
    >>> deflatten({('a', 'b'): 'd', ('a', 'c'): 'e'}, sep=None)
    {'a': {'b': 'd', 'c': 'e'}}
    >>> deflatten({'a.b': 1, 'a': 2})
    Traceback (most recent call last):
      ...
    AssertionError: Conflicting keys! ('a',)
    >>> deflatten({'a': 1, 'a.b': 2})
    Traceback (most recent call last):
      ...
    AssertionError: Conflicting keys! ('a', 'b')

    """
    ret = {}
    if sep is not None:
        d = {
            tuple(k.split(sep, maxdepth)): v for k, v in d.items()
        }

    for keys, v in d.items():
        sub_dict = ret
        for sub_key in keys[:-1]:
            if sub_key not in sub_dict:
                sub_dict[sub_key] = {}
            assert isinstance(sub_dict[sub_key], dict), (
                f'Conflicting keys! {keys}'
            )
            sub_dict = sub_dict[sub_key]
        assert keys[-1] not in sub_dict, f'Conflicting keys! {keys}'
        sub_dict[keys[-1]] = v
    return ret



from sed_scores_eval.base_modules.precision_recall import (
    precision_recall_curve_from_intermediate_statistics,
    fscore_curve_from_intermediate_statistics,
    single_fscore_from_intermediate_statistics,
    best_fscore_from_intermediate_statistics,
    average_precision_from_intermediate_statistics,
)
from sed_scores_eval.segment_based.intermediate_statistics import accumulated_intermediate_statistics


def precision_recall_curve(
        scores, ground_truth, audio_durations, *, deltas=None,
        segment_length=1., time_decimals=6, num_jobs=1,
):
    """Compute segment-based precision-recall curve.

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        audio_durations: The duration of each audio file in the evaluation set.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        segment_length: the segment length of the segments that are to be
            evaluated.
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high detected or ground truth events that have
            onsets or offsets right on a segment boundary may swap over to the
            adjacent segment because of small deviations due to limited
            floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns: (all arrays sorted by corresponding recall)
        precisions ((dict of) 1d np.ndarray): precision values for all operating points
        recalls ((dict of) 1d np.ndarray): recall values for all operating points
        scores ((dict of) 1d np.ndarray): score values that the threshold has to
            fall below to obtain corresponding precision-recall pairs
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        audio_durations=audio_durations, segment_length=segment_length,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    return precision_recall_curve_from_intermediate_statistics(
        intermediate_stats
    )


def fscore_curve(
        scores, ground_truth, audio_durations, *, deltas=None,
        segment_length=1., beta=1., time_decimals=6, num_jobs=1,
):
    """Compute segment-based f-scores with corresponding precisions, recalls
    and intermediate statistics for various operating points

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        audio_durations: The duration of each audio file in the evaluation set.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        segment_length: the segment length of the segments that are to be
            evaluated.
        beta: \beta parameter for f-score computation
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high detected or ground truth events that have
            onsets or offsets right on a segment boundary may swap over to the
            adjacent segment because of small deviations due to limited
            floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns: (all arrays sorted by corresponding score)
        f_beta ((dict of) 1d np.ndarray): f-score values  for all operating
            points
        precisions ((dict of) 1d np.ndarray): precision values for all operating points
        recalls ((dict of) 1d np.ndarray): recall values for all operating points
        scores ((dict of) 1d np.ndarray): score values that the threshold has to
            fall below to obtain corresponding precision-recall pairs
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps': 1d np.ndarray of true positive counts for each score
            'fps': 1d np.ndarray of false positive counts for each score
            'n_ref': integer number of ground truth events

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        audio_durations=audio_durations, segment_length=segment_length,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    return fscore_curve_from_intermediate_statistics(
        intermediate_stats, beta=beta,
    )


def fscore(
        scores, ground_truth, audio_durations, threshold, *, deltas=None,
        segment_length, beta=1., time_decimals=6, num_jobs=1,
):
    """Get segment-based f-score with corresponding precision, recall and
    intermediate statistics for a specific decision threshold

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        audio_durations: The duration of each audio file in the evaluation set.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        threshold ((dict of) float): threshold that is to be evaluated.
        segment_length: the segment length of the segments that are to be
            evaluated.
        beta: \beta parameter for f-score computation
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high detected or ground truth events that have
            onsets or offsets right on a segment boundary may swap over to the
            adjacent segment because of small deviations due to limited
            floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        fscore ((dict of) float): fscore value for threshold
        precision ((dict of) float): precision value for threshold
        recall ((dict of) float): recall value for threshold
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps' (int): true positive count for threshold
            'fps' (int): false positive count for threshold
            'n_ref' (int): number of ground truth events

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        audio_durations=audio_durations, segment_length=segment_length,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    return single_fscore_from_intermediate_statistics(
        intermediate_stats, threshold=threshold, beta=beta,
    )


def best_fscore(
        scores, ground_truth, audio_durations, *, deltas=None,
        segment_length=1., min_precision=0., min_recall=0., beta=1.,
        time_decimals=6, num_jobs=1,
):
    """Get the best possible (macro-averaged) segment-based f-score with
    corresponding precision, recall, intermediate statistics and decision
    threshold

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        audio_durations: The duration of each audio file in the evaluation set.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        segment_length: the segment length of the segments that are to be
            evaluated.
        min_precision: the minimum precision that must be achieved.
        min_recall: the minimum recall that must be achieved. If the
            constraint(s) cannot be achieved at any threshold, however,
            fscore, precision, recall and threshold of 0,1,0,inf are returned.
        beta: \beta parameter for f-score computation
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high detected or ground truth events that have
            onsets or offsets right on a segment boundary may swap over to the
            adjacent segment because of small deviations due to limited
            floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        f_beta ((dict of) float): best achievable f-score value
        precision ((dict of) float): precision value at best fscore
        recall ((dict of) float): recall value at best fscore
        threshold ((dict of) float): threshold to obtain best fscore which is
            centered between the score that the threshold has to fall below
            and the next smaller score which results in different intermediate
            statistics.
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps' (int): true positive count at best fscore
            'fps' (int): false positive count at best fscore
            'n_ref' (int): number of ground truth events

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth, deltas=deltas,
        audio_durations=audio_durations, segment_length=segment_length,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )
    return best_fscore_from_intermediate_statistics(
        intermediate_stats, beta=beta,
        min_precision=min_precision, min_recall=min_recall,
    )


def average_precision(
        scores, ground_truth, audio_durations, *, deltas=None,
        segment_length, time_decimals=6, num_jobs=1,
):
    """Compute average precision

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        audio_durations: The duration of each audio file in the evaluation set.
        deltas (dict of dicts of tuples): Must be deltas as returned by
            `accumulated_intermediate_statistics_from_deltas`. If not provided,
            deltas are computed within this function. Providing deltas is useful
            if deltas are used repeatedly as, e.g., with bootstrapped evaluation,
            to save computing time.
        segment_length: the segment length of the segments that are to be
            evaluated.
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high detected or ground truth events that have
            onsets or offsets right on a segment boundary may swap over to the
            adjacent segment because of small deviations due to limited
            floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        average_precision ((dict of) float): mean and class-wise average
            precisions
        pr_curve: Precision-Recall curve(s) as provided by
            `precision_recall_curve_from_intermediate_statistics`

    """
    intermediate_stats, _ = accumulated_intermediate_statistics(
        scores=scores, ground_truth=ground_truth,
        audio_durations=audio_durations, deltas=deltas,
        segment_length=segment_length, time_decimals=time_decimals,
        num_jobs=num_jobs,
    )
    return average_precision_from_intermediate_statistics(intermediate_stats)



import numpy as np
from sed_scores_eval.base_modules.scores import validate_score_dataframe
from sed_scores_eval.base_modules.ground_truth import event_counts_and_durations
from sed_scores_eval.base_modules import statistics
from sed_scores_eval.base_modules.io import parse_inputs, parse_ground_truth
from sed_scores_eval.base_modules.detection import onset_deltas


def intermediate_statistics_deltas(
        scores, ground_truth, *, dtc_threshold, gtc_threshold,
        cttc_threshold=None, time_decimals=6, num_jobs=1
):
    """Compute intersection-based intermediate statistics deltas for all audio
    files for all event classes and decision thresholds. See [1] for details
    about intersection-based evaluation. See [2] for details about the deltas
    and the joint computation of intermediate statistics for arbitrary decision
    thresholds.

    [1] C. Bilen, G. Ferroni, F. Tuveri, J. Azcarreta and S. Krstulovic,
    "A Framework for the Robust Evaluation of Sound Event Detection",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2020, pp. 61–65

    [2] J.Ebbers, R.Serizel, and R.Haeb-Umbach
    "Threshold-Independent Evaluation of Sound Event Detection Scores",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2022

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        dtc_threshold (float): detection tolerance criterion threshold
        gtc_threshold (float): ground truth intersection criterion threshold
        cttc_threshold (float): cross trigger tolerance criterion threshold
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, e.g., a detection with an ground truth intersection
            exactly matching the DTC, may be falsely counted as false detection
            because of small deviations due to limited floating point precision.
        num_jobs (int): the number of processes to use. Default is 1 in which
            case no multiprocessing is used.

    Returns:
        deltas (dict of dicts of tuples): For each audio clip for each event class:
            change_point_scores (dict of dicts): 1d array of scores at which the
                intermediate statistics change, when the threshold falls below
                it, for that particular clip and class.
            dict of delta values: provides for each intermediate statistic a
                1d array of the delta (change), when the threshold falls below
                each of the change_point_scores.
                Required intermediate statistics deltas:
                 'tps' (1d np.ndarray): true positive counts for each score
                 'fps' (1d np.ndarray): false positive counts for each score
                 'cts' (dict of 1d np.ndarray): cross triggers with each of the
                    other classes for each score

    """
    scores, ground_truth, keys = parse_inputs(scores, ground_truth)
    _, event_classes = validate_score_dataframe(
        scores[keys[0]])
    return statistics.intermediate_statistics_deltas(
        scores, ground_truth,
        intermediate_statistics_fn=statistics_fn,
        acceleration_fn=acceleration_fn,
        dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
        cttc_threshold=cttc_threshold,
        time_decimals=time_decimals, num_jobs=num_jobs,
    )


def accumulated_intermediate_statistics_from_deltas(deltas, ground_truth):
    """Compute intersection-based intermediate statistics over all audio files
    for all event classes and decision thresholds. See [1] for details about
    intersection-based evaluation. See [2] for details about the joint
    computation of intermediate statistics for arbitrary decision thresholds.

    [1] C. Bilen, G. Ferroni, F. Tuveri, J. Azcarreta and S. Krstulovic,
    "A Framework for the Robust Evaluation of Sound Event Detection",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2020, pp. 61–65

    [2] J.Ebbers, R.Serizel, and R.Haeb-Umbach
    "Threshold-Independent Evaluation of Sound Event Detection Scores",
    in Proc. IEEE International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
    2022

    Args:
        deltas (dict of dicts of tuples): For each audio clip for each event class:
            change_point_scores (dict of dicts): 1d array of scores at which the
                intermediate statistics change, when the threshold falls below
                it, for that particular clip and class.
            dict of delta values: provides for each intermediate statistic a
                1d array of the delta (change), when the threshold falls below
                each of the change_point_scores.
                Required intermediate statistics deltas:
                 'tps' (1d np.ndarray): true positive counts for each score
                 'fps' (1d np.ndarray): false positive counts for each score
                 'cts' (dict of 1d np.ndarray): cross triggers with each of the
                    other classes for each score
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.

    Returns (dict of tuples): for each event class a tuple of 1d scores array
        and a dict of intermediate statistics with the following keys
        (where each array has the same length as the scores array):
            "tps": true positives count array
            "fps": false positives count array
            "cts": dict of cross trigger count arrays with each other class
            "n_ref": integer number of target class ground truth events
            "t_ref": combined duration of all target class ground truth events
            "n_ref_other": dict of integer numbers of ground truth events from
                each other class
            "t_ref_other": dict of combined durations of ground truth events
                from each other class

    """
    multi_label_statistics = statistics.accumulated_intermediate_statistics_from_deltas(deltas)
    event_classes = multi_label_statistics.keys()
    n_ref, t_ref = event_counts_and_durations(
        ground_truth, event_classes=event_classes
    )
    return {
        class_name: (
            cp_scores_cls,
            {
                **stats_cls,
                'n_ref': n_ref[class_name],
                't_ref': t_ref[class_name],
                'n_ref_other': {
                    ocls: n_ref[ocls]
                    for ocls in event_classes if ocls != class_name
                },
                't_ref_other': {
                    ocls: t_ref[ocls]
                    for ocls in event_classes if ocls != class_name
                },
            },
        )
        for class_name, (cp_scores_cls, stats_cls) in multi_label_statistics.items()
    }


def accumulated_intermediate_statistics(
        scores, ground_truth, *, deltas=None, dtc_threshold, gtc_threshold,
        cttc_threshold=None, time_decimals=6, num_jobs=1
):
    """Cascade of `intermediate_statistics_deltas` and `accumulated_intermediate_statistics_from_deltas`.

    Args:
        scores:
        ground_truth:
        deltas:
        dtc_threshold:
        gtc_threshold:
        cttc_threshold:
        time_decimals:
        num_jobs:

    Returns:

    """
    if deltas is None:
        scores, ground_truth, audio_ids = parse_inputs(scores, ground_truth)
        deltas = intermediate_statistics_deltas(
            scores=scores, ground_truth=ground_truth,
            dtc_threshold=dtc_threshold, gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold,
            time_decimals=time_decimals, num_jobs=num_jobs,
        )
    else:
        audio_ids = list(deltas.keys())
        ground_truth = parse_ground_truth(
            ground_truth, audio_ids=audio_ids, additional_ids_ok=True)
    return accumulated_intermediate_statistics_from_deltas(deltas, ground_truth), audio_ids


def statistics_fn(
        detection_onset_times, detection_offset_times,
        target_onset_times, target_offset_times,
        other_onset_times, other_offset_times,
        dtc_threshold, gtc_threshold, cttc_threshold,
        time_decimals=6,
):
    """Compute intersection-based intermediate statistics for a single audio
    and single target class given detected onset/offset times, target class
    ground truth onset/offset times and other classes' ground truth
    onset/offset times.

    Args:
        detection_onset_times (np.ndarray): (B, M) onset times of detected
            target class events with M being the number of detected target
            class events, and B being an independent dimension.
        detection_offset_times (np.ndarray): (B, M) offset times of detected
            target class events with M being the number of detected target
            class events, and B being an independent dimension. Note that it
            may include offset times which are equal to the corresponding onset
            time, which indicates that the event is inactive at that specific
            position b along the independent axis and must not be counted as a
            detection.
        target_onset_times (1d np.ndarray): onset times of target class ground
            truth events.
        target_offset_times (1d np.ndarray): offset times of target class
            ground truth events.
        other_onset_times (dict of 1d np.ndarrays): onset times of other class
            ground truth events
        other_offset_times (dict of 1d np.ndarrays): offset times of other
            class ground truth events
        dtc_threshold (float): detection tolerance criterion threshold
        gtc_threshold (float): ground truth intersection criterion threshold
        cttc_threshold (float): cross trigger tolerance criterion threshold
        time_decimals (int): the decimal precision used for evaluation. If
            chosen to high, e.g., a detection with a ground truth intersection
            exactly matching the DTC, may be falsely counted as false detection
            because of small deviations due to limited floating point precision.

    Returns (dict of 1d np.ndarrays): dict of intermediate statistics with the
        following keys (where each array has the length B):
            "tps": true positives count array
            "fps": false positives count array
            "cts": dict of cross trigger count arrays with each other class

    """
    det_crit = detection_offset_times > detection_onset_times
    num_detections = det_crit.sum(-1)
    ground_truth_intersections = np.maximum(
        np.minimum(detection_offset_times[..., None], target_offset_times)
        - np.maximum(detection_onset_times[..., None], target_onset_times),
        0.,
    )
    total_intersection_with_gt_events = np.round(
        np.sum(ground_truth_intersections, axis=-1), decimals=time_decimals)
    detection_lengths = detection_offset_times - detection_onset_times
    dtc = (
        total_intersection_with_gt_events
        >= np.round(dtc_threshold * detection_lengths, decimals=time_decimals)
    ) * det_crit
    num_relevant_detections = dtc.sum(-1)
    fps = num_detections - num_relevant_detections
    total_intersection_with_relevant_detections = np.round(
        np.sum(dtc[..., None] * ground_truth_intersections, axis=-2),
        decimals=time_decimals
    )
    gt_lengths = target_offset_times - target_onset_times
    gtc = (
        total_intersection_with_relevant_detections
        >= np.round(gtc_threshold * gt_lengths, decimals=time_decimals)
    )
    tps = gtc.sum(-1)
    if cttc_threshold is None:
        cts = np.zeros_like(tps)
    else:
        cts = {}
        assert other_onset_times.keys() == other_offset_times.keys(), (other_onset_times.keys(), other_offset_times.keys())
        for other_class_name in other_onset_times.keys():
            gt_onset_times = other_onset_times[other_class_name]
            gt_offset_times = other_offset_times[other_class_name]
            assert len(gt_onset_times) == len(gt_offset_times)
            if len(gt_onset_times) == 0:
                cts[other_class_name] = np.zeros_like(tps)
            else:
                other_class_intersections = np.maximum(
                    np.minimum(detection_offset_times[..., None], gt_offset_times)
                    - np.maximum(detection_onset_times[..., None], gt_onset_times),
                    0.,
                )
                total_intersection_with_other_gt_events = np.round(
                    np.sum((1-dtc[..., None]) * other_class_intersections, axis=-1),
                    decimals=time_decimals
                )
                cttc = (
                    total_intersection_with_other_gt_events
                    >= np.round(cttc_threshold * detection_lengths, decimals=time_decimals)
                ) * det_crit
                cts[other_class_name] = cttc.sum(-1)
    return {'tps': tps, 'fps': fps, 'cts': cts}


def acceleration_fn(
        scores, timestamps,
        target_onset_times, target_offset_times,
        other_onset_times, other_offset_times,
        dtc_threshold, gtc_threshold, cttc_threshold,
        time_decimals=6,
):
    # return np.unique(scores), None, None
    onset_deltas_ = onset_deltas(scores)
    change_points = np.abs(onset_deltas_) > .5
    if (len(target_onset_times) == 0) and (
        (cttc_threshold is None) or all([
            len(gt_onset_times) == 0 for gt_onset_times in other_onset_times
        ])
    ):
        cp_scores = scores[change_points]
        if cttc_threshold is None:
            cts_deltas = np.zeros_like(cp_scores)
        else:
            cts_deltas = {
                other_class_name: np.zeros_like(cp_scores)
                for other_class_name in other_onset_times.keys()
            }
        deltas = {
            'fps': onset_deltas_[change_points],
            'tps': np.zeros_like(cp_scores),
            'cts': cts_deltas,
        }
        return None, cp_scores, deltas
    return None, None, None
    # if cttc_threshold is None:
    #     gt_onset_times = target_onset_times
    #     gt_offset_times = target_offset_times
    # else:
    #     other_class_names = list(other_onset_times.keys())
    #     gt_onset_times = np.concatenate([target_onset_times] + [other_onset_times[ocls] for ocls in other_class_names])
    #     gt_offset_times = np.concatenate([target_offset_times] + [other_offset_times[ocls] for ocls in other_class_names])
    #
    # for onset_time, offset_time in zip(gt_onset_times, gt_offset_times):
    #     onset_idx, offset_idx = onset_offset_times_to_score_indices(onset_time, offset_time, timestamps)
    #     change_points[onset_idx:offset_idx] = True
    #     right_sided_cummin_indices = offset_idx-1 + cummin(scores[offset_idx-1:])[1]
    #     change_points[right_sided_cummin_indices] = True
    #     left_sided_cummin_indices = onset_idx - cummin(scores[:onset_idx+1][::-1])[1]
    #     change_points[left_sided_cummin_indices] = True
    #
    # return np.unique(scores[change_points]), None, None



import numpy as np
from sed_scores_eval.base_modules.scores import validate_score_dataframe
from sed_scores_eval.base_modules.io import parse_inputs


def metrics(scores, ground_truth, threshold):
    """Reference metrics implementation using the sed_eval package
    (https://tut-arg.github.io/sed_eval/), which, however, does not allow to
    compute metrics at various operating points simultaneously.
    This function is primarily used for testing purposes.

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            tags for each audio clip or a file path from where the ground truth
            can be loaded.
        threshold ((dict of) float): threshold that is to be evaluated.

    Returns (sed_eval.audio_tag.AudioTaggingMetrics): audio tagging metrics

    """
    import sed_eval
    import dcase_util

    scores, ground_truth, keys = parse_inputs(scores, ground_truth, tagging=True)
    _, event_classes = validate_score_dataframe(scores[keys[0]])
    if isinstance(threshold, dict):
        threshold = [threshold[event_class] for event_class in event_classes]
        if not all([np.isscalar(thr) for thr in threshold]):
            raise ValueError('All values of thresholds dict must be scalars')
        threshold = np.asanyarray(threshold)
    elif not np.isscalar(threshold):
        raise ValueError(
            f'threshold must be (dict of) scalar(s) but {type(threshold)} '
            f'was given.'
        )

    estimated_tag_list = []
    for audio_id in keys:
        scores_k = scores[audio_id]
        validate_score_dataframe(scores_k, event_classes=event_classes)
        scores_k = scores_k[event_classes].to_numpy().max(0)
        estimated_tags = scores_k > threshold
        estimated_tags = [
            class_name for c, class_name in enumerate(event_classes)
            if estimated_tags[c]
        ]
        estimated_tag_list.append({
            'filename': f'{audio_id}.wav',
            'tags': ','.join(estimated_tags),
        })

    estimated_tag_list = dcase_util.containers.MetaDataContainer(
        estimated_tag_list)
    reference_tag_list = dcase_util.containers.MetaDataContainer([
        {
            'filename': f'{audio_id}.wav',
            'tags': ','.join(tags),
        }
        for audio_id, tags in ground_truth.items()
    ])

    tag_evaluator = sed_eval.audio_tag.AudioTaggingMetrics(
        tags=reference_tag_list.unique_tags
    )
    tag_evaluator.evaluate(
        reference_tag_list=reference_tag_list,
        estimated_tag_list=estimated_tag_list,
    )
    return tag_evaluator



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

import setuptools as st
from codecs import open
from os import path
import numpy as np
from Cython.Build import cythonize


ext_modules = cythonize(
    [
        'sed_scores_eval/base_modules/cy_detection.pyx',
        'sed_scores_eval/base_modules/cy_medfilt.pyx',
     ]
)

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

st.setup(
    name='sed_scores_eval',
    version='0.0.3',
    description='(Threshold-Independent) Evaluation of Sound Event Detection Scores',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/fgnt/sed_scores_eval',
    author='Department of Communications Engineering, Paderborn University',
    author_email='sek@nt.upb.de',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    keywords='sound recognition evaluation from classification scores',
    packages=st.find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=[
        'numpy',
        'pandas',
        'scipy',
        'pathlib',
        'lazy_dataset',
        'einops',
        'sed_eval',
        'Cython',
    ],
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage', 'jupyter', 'matplotlib'],
    },
    ext_modules=ext_modules,
    package_data={'sed_scores_eval': ['**/*.pyx']},  # https://stackoverflow.com/a/60751886
    include_dirs=[np.get_include()],
)



import pandas as pd
from pathlib import Path
from sed_scores_eval.base_modules.io import parse_inputs, write_detection


def metrics(
        scores, ground_truth, threshold, *, segment_length=1,
):
    """Reference metrics implementation using the sed_eval package
    (https://tut-arg.github.io/sed_eval/), which, however, does not allow to
    compute metrics at various operating points simultaneously.
    This function is primarily used for testing purposes.

    Args:
        scores (dict, str, pathlib.Path): dict of SED score DataFrames
            (cf. sed_scores_eval.utils.scores.create_score_dataframe)
            or a directory path (as str or pathlib.Path) from where the SED
            scores can be loaded.
        ground_truth (dict, str or pathlib.Path): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio clip or a
            file path from where the ground truth can be loaded.
        threshold ((dict of) float): threshold that is to be evaluated.
        segment_length: the segment length of the segments that are to be
            evaluated.

    Returns (sed_eval.sound_event.SegmentBasedMetrics): segment-based metrics

    """
    import tempfile
    import sed_eval
    import dcase_util

    assert isinstance(ground_truth, (str, Path)), type(ground_truth)
    ground_truth = str(ground_truth)
    audio_format = pd.read_csv(ground_truth, sep="\t")['filename'][0].split('.')[-1]
    scores, *_ = parse_inputs(scores, ground_truth)
    tmp_fid = tempfile.NamedTemporaryFile(delete=False)
    tmp_filepath = Path(tmp_fid.name)
    tmp_fid.close()
    try:
        write_detection(
            scores=scores, threshold=threshold,
            filepath=tmp_filepath, audio_format=audio_format,
        )
        reference_event_list = sed_eval.io.load_event_list(
            filename=ground_truth, file_format='CSV',
        )
        reference_event_list = dcase_util.containers.MetaDataContainer(
            [entry for entry in reference_event_list if
             entry['event_label'] is not None]
        )
        estimated_event_list = sed_eval.io.load_event_list(
            filename=str(tmp_filepath), file_format='CSV',
        )

        all_data = dcase_util.containers.MetaDataContainer()
        all_data += reference_event_list
        all_data += estimated_event_list

        event_labels = all_data.unique_event_labels

        # Start evaluating
        # Create metrics classes, define parameters
        segment_based_metrics = sed_eval.sound_event.SegmentBasedMetrics(
            event_label_list=event_labels,
            time_resolution=segment_length,
        )

        # Go through files
        for filename in all_data.unique_files:
            reference_event_list_for_current_file = reference_event_list.filter(
                filename=filename
            )

            estimated_event_list_for_current_file = estimated_event_list.filter(
                filename=filename
            )
            segment_based_metrics.evaluate(
                reference_event_list=reference_event_list_for_current_file,
                estimated_event_list=estimated_event_list_for_current_file
            )
        return segment_based_metrics
    finally:
        if tmp_filepath.exists():
            tmp_filepath.unlink()



import numpy as np
from sed_scores_eval.utils import parallel


def bootstrap_from_deltas(
        metric_fn, deltas, *,
        n_folds=5, n_iterations=20, num_jobs=1,
        **metric_fn_kwargs,
):
    if isinstance(deltas, (list, tuple)):
        audio_ids = sorted(deltas[0].keys())
    else:
        audio_ids = sorted(deltas.keys())
    split_indices = np.linspace(0, len(audio_ids), n_folds+1).astype(int)
    audio_id_fractions = []
    for i in range(n_iterations):
        np.random.RandomState(i).shuffle(audio_ids)
        for j in range(n_folds):
            audio_id_fractions.append(list(
                audio_ids[:split_indices[j]] + audio_ids[split_indices[j+1]:]
            ))

    if isinstance(deltas, (list, tuple)):
        deltas_fractions = [
            [
                {audio_id: delts[audio_id] for audio_id in audio_id_fraction}
                for delts in deltas
            ]
            for audio_id_fraction in audio_id_fractions
        ]
    else:
        deltas_fractions = [
            {audio_id: deltas[audio_id] for audio_id in audio_id_fraction}
            for audio_id_fraction in audio_id_fractions
        ]
    return list(zip(*parallel.map(
        deltas_fractions, arg_keys='deltas',
        func=metric_fn, max_jobs=num_jobs,
        **metric_fn_kwargs,
    )))


def confidence_interval(bootstrapped_outputs, confidence=.9):
    if isinstance(bootstrapped_outputs[0], dict):
        mean_low_high = {
            class_name: confidence_interval([
                output[class_name] for output in bootstrapped_outputs
            ])
            for class_name in bootstrapped_outputs[0]
        }
        return mean_low_high

    mean = np.mean(bootstrapped_outputs)
    low = np.percentile(
        bootstrapped_outputs, ((1 - confidence) / 2) * 100
    )
    high = np.percentile(
        bootstrapped_outputs, (confidence + ((1 - confidence) / 2)) * 100
    )
    return float(mean), float(low), float(high)




import numpy as np
from sed_scores_eval.base_modules.curves import get_curve_idx_for_threshold, xsort


def error_rate_curve_from_intermediate_statistics(
        scores_intermediate_statistics
):
    """Compute error rate curve from intermediate_statistics

    Args:
        scores_intermediate_statistics ((dict of) tuple): tuple of scores array
            and dict of intermediate_statistics with the following key value
            pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events
            If dict input is provided, keys are expected to be class names with
            corresponding scores/intermediate_statistics as values.

    Returns: (all arrays sorted by corresponding recall)
        error_rates ((dict of) 1d np.ndarray): error rates for all operating points
        insertion_rate ((dict of) 1d np.ndarray): insertion rates for all operating points
        deletion_rates ((dict of) 1d np.ndarray): deletion rates for all operating points
        scores ((dict of) 1d np.ndarray): score values that the threshold has to
            fall below to obtain corresponding error rate.
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events

    """
    if isinstance(scores_intermediate_statistics, dict):
        error_rate_curves = {}
        for class_name, scores_stats in scores_intermediate_statistics.items():
            error_rate_curves[class_name] = error_rate_curve_from_intermediate_statistics(scores_stats)
        return error_rate_curves

    if not isinstance(scores_intermediate_statistics, (list, tuple)):
        raise ValueError(
            f'scores_intermediate_statistics must be list/tuple of length 2, '
            f'but {type(scores_intermediate_statistics)} was given.'
        )
    if len(scores_intermediate_statistics) != 2:
        raise ValueError(
            f'scores_intermediate_statistics must be list/tuple of length 2, '
            f'but list/tuple of length {len(scores_intermediate_statistics)}'
            f'was given.'
        )
    scores, stats = scores_intermediate_statistics
    if not isinstance(stats, dict):
        raise ValueError(
            f'intermediate_statistics must be dict but {type(stats)} was given.')
    if not all([key in stats for key in ['tps', 'fps', 'n_ref']]):
        raise ValueError(
            f'intermediate_statistics must contain keys "tps", "fps" and "n_ref". '
            f'Provided keys: {sorted(stats.keys())}.'
        )
    insertions = stats['fps']
    deletions = stats['n_ref'] - stats['tps']
    errors = insertions + deletions
    er = errors / stats['n_ref']
    ir = insertions / stats['n_ref']
    dr = deletions / stats['n_ref']
    er, scores, ir, dr, stats = xsort(er, scores, ir, dr, stats)
    return er, ir, dr, scores, stats


def best_error_rate_from_intermediate_statistics(
        scores_intermediate_statistics,
        max_insertion_rate=None, max_deletion_rate=None,
):
    """Get the best possible (macro-average) error rate with corresponding
    insertion rate, deletion_rate, intermediate statistics and decision
    threshold

    Args:
        scores_intermediate_statistics ((dict of) tuple): tuple of scores array
            and dict of intermediate_statistics with the following key value
            pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events
            If dict input is provided keys are expected to be class names with
            corresponding scores/intermediate_statistics as values.
        max_insertion_rate: a maximal insertion rate that must not be exceeded.
        max_deletion_rate: a maximal deletion rate that must not be exceeded.
            If the constraint(s) cannot be achieved at any threshold, however,
            er, ir, dr and threshold of 1,0,1,inf are returned.

    Returns:
        error_rate ((dict of) float): best achievable error_rate value
        insertion_rate ((dict of) float): insertion rate at best error rate
        deletion_rate ((dict of) float): deletion rate at best error rate
        threshold ((dict of) float): threshold to obtain best error rate which
            is centered between the score that the threshold has to fall below
            and the next smaller score which results in different intermediate
            statistics.
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps' (int): true positive count at best error rate
            'fps' (int): false positive count at best error rate
            'n_ref' (int): number of ground truth events

    """
    if isinstance(scores_intermediate_statistics, dict):
        er, ir, dr, thresholds, intermediate_stats = {}, {}, {}, {}, {}
        for class_name, scores_stats in scores_intermediate_statistics.items():
            (
                er[class_name], ir[class_name], dr[class_name],
                thresholds[class_name], intermediate_stats[class_name]
            ) = best_error_rate_from_intermediate_statistics(
                scores_stats,
                max_insertion_rate=max_insertion_rate,
                max_deletion_rate=max_deletion_rate,
            )
        er['macro_average'] = np.mean([er[class_name] for class_name in er])
        ir['macro_average'] = np.mean([ir[class_name] for class_name in ir])
        dr['macro_average'] = np.mean([dr[class_name] for class_name in dr])
        return er, ir, dr, thresholds, intermediate_stats
    er, ir, dr, scores, intermediate_stats = \
        error_rate_curve_from_intermediate_statistics(
            scores_intermediate_statistics,
        )
    if max_insertion_rate is not None:
        er[ir < max_insertion_rate] = np.inf
    if max_deletion_rate is not None:
        er[dr < max_insertion_rate] = np.inf
    best_idx = len(er) - 1 - np.argmin(er[::-1], axis=0)
    threshold = (
        (scores[best_idx] + scores[best_idx-1])/2 if best_idx > 0 else -np.inf
    )

    def _recursive_get_item(stats, idx):
        if isinstance(stats, dict):
            return {key: _recursive_get_item(stats[key], idx) for key in stats}
        if np.isscalar(stats):
            return stats
        return stats[idx]
    return (
        er[best_idx], ir[best_idx], dr[best_idx], threshold,
       _recursive_get_item(intermediate_stats, best_idx)
    )


def single_error_rate_from_intermediate_statistics(
        scores_intermediate_statistics, threshold):
    """Get error rate with corresponding insertion rate, deletion rate and
    intermediate statistics for a specific decision threshold

    Args:
        scores_intermediate_statistics ((dict of) tuple): tuple of scores array
            and dict of intermediate_statistics with the following key value
            pairs:
             'tps' (1d np.ndarray): true positive counts for each score
             'fps' (1d np.ndarray): false positive counts for each score
             'n_ref' (int): number of ground truth events
            If dict input is provided keys are expected to be class names with
            corresponding scores/intermediate_statistics as values.
        threshold ((dict of) float): threshold that is to be evaluated.

    Returns:
        error_rate ((dict of) float): error rate for threshold
        insertion_rate ((dict of) float): insertion rate for threshold
        deletion rate ((dict of) float): deletion rate for threshold
        intermediate_statistics ((dict of) dict): dict of
            intermediate_statistics with the following key value pairs:
            'tps' (int): true positive count for threshold
            'fps' (int): false positive count for threshold
            'n_ref' (int): number of ground truth events

    """
    if isinstance(threshold, dict):
        if not all([np.isscalar(thr) for thr in threshold]):
            raise ValueError('All values of thresholds dict must be scalars')
    elif not np.isscalar(threshold):
        raise ValueError(
            f'threshold must be (dict of) scalar(s) but '
            f'{type(threshold)} was given.'
        )
    if isinstance(scores_intermediate_statistics, dict):
        er, ir, dr, intermediate_stats = {}, {}, {}, {}
        for class_name, scores_stats in scores_intermediate_statistics.items():
            (
                er[class_name], ir[class_name], dr[class_name],
                intermediate_stats[class_name]
            ) = single_error_rate_from_intermediate_statistics(
                scores_stats,
                threshold[class_name]
                if isinstance(threshold, dict)
                else threshold,
            )
        er['macro_average'] = np.mean([er[class_name] for class_name in er])
        ir['macro_average'] = np.mean([ir[class_name] for class_name in ir])
        dr['macro_average'] = np.mean([dr[class_name] for class_name in dr])
        return er, ir, dr, intermediate_stats
    error_rate_curve = error_rate_curve_from_intermediate_statistics(
        scores_intermediate_statistics
    )
    er, ir, dr, scores, intermediate_stats = error_rate_curve
    return _single_er_from_er_curve(
        er, ir, dr, scores, threshold,
        intermediate_statistics=intermediate_stats
    )


def _single_er_from_er_curve(
        error_rates, insertion_rate, deletion_rate, scores, threshold,
        intermediate_statistics=None
):
    idx = get_curve_idx_for_threshold(scores, threshold)
    error_rate = error_rates[idx]
    insertion_rate = insertion_rate[idx]
    deletion_rate = deletion_rate[idx]
    if intermediate_statistics is None:
        return error_rate, insertion_rate, deletion_rate
    else:
        return error_rate, insertion_rate, deletion_rate, {
            key: stat if np.isscalar(stat) else stat[idx]
            for key, stat in intermediate_statistics.items()
        }


def error_rate_from_sed_eval_metrics(sed_eval_metrics):
    """extract class-wise and macro-averaged error,insertion and deletion rates
    from sed_eval metrics object

    Args:
        sed_eval_metrics:

    Returns:
        er (dict of float): error rates
        ir (dict of float): insertion rates
        dr (dict of float): deletion rates

    """
    er = {}
    ir = {}
    dr = {}
    sed_eval_results_classwise = sed_eval_metrics.results_class_wise_metrics()
    for key in sed_eval_results_classwise:
        er[key] = sed_eval_results_classwise[key]['error_rate']['error_rate']
        ir[key] = sed_eval_results_classwise[key]['error_rate']['insertion_rate']
        dr[key] = sed_eval_results_classwise[key]['error_rate']['deletion_rate']
    sed_eval_results_macro = sed_eval_metrics.results_class_wise_average_metrics()
    er['macro_average'] = sed_eval_results_macro['error_rate']['error_rate']
    ir['macro_average'] = sed_eval_results_macro['error_rate']['insertion_rate']
    dr['macro_average'] = sed_eval_results_macro['error_rate']['deletion_rate']
    sed_eval_results_micro = sed_eval_metrics.results_overall_metrics()
    er['micro_average'] = sed_eval_results_micro['error_rate']['error_rate']
    ir['micro_average'] = sed_eval_results_micro['error_rate']['insertion_rate']
    dr['micro_average'] = sed_eval_results_micro['error_rate']['deletion_rate']
    return er, ir, dr



from .intermediate_statistics import accumulated_intermediate_statistics, intermediate_statistics_deltas
from .psds import (
    psd_roc, psds, bootstrapped_psds,
    psds_from_psd_roc, multi_class_psd_roc_from_single_class_psd_rocs,
)
from .pipsds import (
    postprocessing_independent_psd_roc, postprocessing_independent_psds,
    bootstrapped_postprocessing_independent_psds, deltas_postprocessing,
    median_filter_independent_psds, bootstrapped_median_filter_independent_psds,
)
from .precision_recall import precision_recall_curve, fscore_curve, fscore, best_fscore
from .error_rate import error_rate_curve, error_rate, best_error_rate
from . import reference
